#!/usr/bin/env python
# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2015  Jon Lund Steffensen <jon_steffensen@uri.edu>

import unittest

from psamm.metabolicmodel import MetabolicModel, FlipableModelView
from psamm.database import DictDatabase
from psamm.reaction import Compound
from psamm.datasource.modelseed import parse_reaction


class TestMetabolicModel(unittest.TestCase):
    def setUp(self):
        # TODO use mock database instead of actual database
        self.database = DictDatabase()
        self.database.set_reaction('rxn_1', parse_reaction('=> (2) |A|'))
        self.database.set_reaction('rxn_2', parse_reaction('|A| <=> |B|'))
        self.database.set_reaction('rxn_3', parse_reaction('|A| => |D[e]|'))
        self.database.set_reaction('rxn_4', parse_reaction('|A| => |C|'))
        self.database.set_reaction('rxn_5', parse_reaction('|C| => |D[e]|'))
        self.database.set_reaction('rxn_6', parse_reaction('|D[e]| =>'))
        self.model = MetabolicModel.load_model(
            self.database, self.database.reactions)

    def test_database_property(self):
        self.assertIs(self.model.database, self.database)

    def test_reaction_set(self):
        self.assertEqual(set(self.model.reactions),
                         {'rxn_1', 'rxn_2', 'rxn_3', 'rxn_4',
                          'rxn_5', 'rxn_6'})

    def test_compound_set(self):
        self.assertEqual(set(self.model.compounds),
                        {Compound('A'), Compound('B'),
                         Compound('C'), Compound('D', 'e')})

    def test_compartments(self):
        self.assertEqual(set(self.model.compartments), {None, 'e'})

    def test_add_reaction_new(self):
        self.database.set_reaction('rxn_7', parse_reaction('|D[e]| => |E[e]|'))
        self.model.add_reaction('rxn_7')
        self.assertIn('rxn_7', set(self.model.reactions))
        self.assertIn(Compound('E', 'e'), set(self.model.compounds))

    def test_add_reaction_existing(self):
        self.model.add_reaction('rxn_1')
        self.assertEqual(
            set(self.model.reactions),
            {'rxn_1', 'rxn_2', 'rxn_3', 'rxn_4', 'rxn_5', 'rxn_6'})
        self.assertEqual(
            set(self.model.compounds),
            {Compound('A'), Compound('B'), Compound('C'), Compound('D', 'e')})

    def test_add_reaction_invalid(self):
        with self.assertRaises(Exception):
            self.model.add_reaction('rxn_7')

    def test_remove_reaction_existing(self):
        self.model.remove_reaction('rxn_2')
        self.assertEqual(
            set(self.model.reactions),
            {'rxn_1', 'rxn_3', 'rxn_4', 'rxn_5', 'rxn_6'})
        self.assertEqual(
            set(self.model.compounds),
            {Compound('A'), Compound('C'), Compound('D', 'e')})

    def test_is_reversible_on_reversible(self):
        self.assertTrue(self.model.is_reversible('rxn_2'))

    def test_is_reversible_on_irreversible(self):
        self.assertFalse(self.model.is_reversible('rxn_1'))
        self.assertFalse(self.model.is_reversible('rxn_3'))

    def test_is_exchange_on_exchange(self):
        self.assertTrue(self.model.is_exchange('rxn_1'))
        self.assertTrue(self.model.is_exchange('rxn_6'))

    def test_is_exchange_on_internal(self):
        self.assertFalse(self.model.is_exchange('rxn_2'))
        self.assertFalse(self.model.is_exchange('rxn_5'))

    def test_add_all_database_reactions(self):
        self.database.set_reaction('rxn_7', parse_reaction('|D| => |E|'))
        added = self.model.add_all_database_reactions()
        self.assertEqual(added, { 'rxn_7' })
        self.assertEqual(set(self.model.reactions), { 'rxn_1', 'rxn_2', 'rxn_3', 'rxn_4', 'rxn_5', 'rxn_6', 'rxn_7' })

    def test_add_all_database_reactions_none(self):
        added = self.model.add_all_database_reactions()
        self.assertEqual(added, set())
        self.assertEqual(set(self.model.reactions), { 'rxn_1', 'rxn_2', 'rxn_3', 'rxn_4', 'rxn_5', 'rxn_6' })

    def test_limits_get_item(self):
        self.assertEqual(self.model.limits['rxn_1'].bounds, (0, 1000))
        self.assertEqual(self.model.limits['rxn_2'].bounds, (-1000, 1000))
        self.assertEqual(self.model.limits['rxn_3'].bounds, (0, 1000))

    def test_limits_get_item_invalid_key(self):
        with self.assertRaises(KeyError):
            a = self.model.limits['rxn_7']

    def test_limits_set_item_is_invalid(self):
        with self.assertRaises(TypeError):
            self.model.limits['rxn_1'] = None

    def test_limits_set_upper_flux_bounds(self):
        self.model.limits['rxn_1'].upper = 500
        self.assertEqual(self.model.limits['rxn_1'].bounds, (0, 500))

    def test_limits_set_upper_flux_bounds_to_invalid(self):
        with self.assertRaises(ValueError):
            self.model.limits['rxn_1'].upper = -20

    def test_limits_delete_upper_flux_bounds(self):
        self.model.limits['rxn_1'].upper = 500
        del self.model.limits['rxn_1'].upper
        self.assertEqual(self.model.limits['rxn_1'].bounds, (0, 1000))

    def test_limits_delete_upper_not_set(self):
        del self.model.limits['rxn_1'].upper
        self.assertEqual(self.model.limits['rxn_1'].bounds, (0, 1000))

    def test_limits_set_lower_flux_bounds(self):
        self.model.limits['rxn_1'].lower = 500
        self.assertEqual(self.model.limits['rxn_1'].bounds, (500, 1000))

    def test_limits_set_lower_flux_bounds_to_invalid(self):
        with self.assertRaises(ValueError):
            self.model.limits['rxn_1'].lower = 1001

    def test_limits_delete_lower_flux_bounds(self):
        self.model.limits['rxn_1'].lower = 500
        del self.model.limits['rxn_1'].lower
        self.assertEqual(self.model.limits['rxn_1'].bounds, (0, 1000))

    def test_limits_delete_lower_not_set(self):
        del self.model.limits['rxn_1'].lower
        self.assertEqual(self.model.limits['rxn_1'].bounds, (0, 1000))

    def test_limits_set_both_flux_bounds(self):
        self.model.limits['rxn_2'].bounds = 1001, 1002
        self.assertEqual(self.model.limits['rxn_2'].lower, 1001)
        self.assertEqual(self.model.limits['rxn_2'].upper, 1002)
        self.assertEqual(self.model.limits['rxn_2'].bounds, (1001, 1002))

    def test_limits_set_both_flux_bounds_to_invalid(self):
        with self.assertRaises(ValueError):
            self.model.limits['rxn_2'].bounds = 10, -1000

    def test_limits_delete_both_flux_bounds(self):
        self.model.limits['rxn_2'].bounds = -10, 800
        del self.model.limits['rxn_2'].bounds
        self.assertEqual(self.model.limits['rxn_2'].bounds, (-1000, 1000))

    def test_limits_set_both_flux_bounds_delete_lower(self):
        self.model.limits['rxn_2'].bounds = -10, 800
        del self.model.limits['rxn_2'].lower
        self.assertEqual(self.model.limits['rxn_2'].bounds, (-1000, 800))

    def test_limits_set_both_flux_bounds_delete_upper(self):
        self.model.limits['rxn_2'].bounds = -10, 800
        del self.model.limits['rxn_2'].upper
        self.assertEqual(self.model.limits['rxn_2'].bounds, (-10, 1000))

    def test_limits_iter(self):
        self.assertEqual(set(iter(self.model.limits)), { 'rxn_1', 'rxn_2', 'rxn_3', 'rxn_4', 'rxn_5', 'rxn_6' })

    def test_limits_len(self):
        self.assertEqual(len(self.model.limits), 6)

class TestMetabolicModelFlipableView(unittest.TestCase):
    def setUp(self):
        # TODO use mock database instead of actual database
        self.database = DictDatabase()
        self.database.set_reaction('rxn_1', parse_reaction('=> (2) |A|'))
        self.database.set_reaction('rxn_2', parse_reaction('|A| <=> |B|'))
        self.database.set_reaction('rxn_3', parse_reaction('|A| => |D|'))
        self.database.set_reaction('rxn_4', parse_reaction('|A| => |C|'))
        self.database.set_reaction('rxn_5', parse_reaction('|C| => |D|'))
        self.database.set_reaction('rxn_6', parse_reaction('|D| =>'))

        model = MetabolicModel.load_model(self.database, self.database.reactions)
        self.model = FlipableModelView(model)

    def test_flipable_model_view_matrix_get_item_after_flip(self):
        self.model.flip({ 'rxn_4' })
        self.assertEqual(self.model.matrix[Compound('A'), 'rxn_1'], 2)
        self.assertEqual(self.model.matrix[Compound('A'), 'rxn_2'], -1)
        self.assertEqual(self.model.matrix[Compound('A'), 'rxn_4'], 1)
        self.assertEqual(self.model.matrix[Compound('C'), 'rxn_4'], -1)

    def test_flipable_model_view_matrix_get_item_after_double_flip(self):
        self.model.flip({ 'rxn_4', 'rxn_5' })
        self.model.flip({ 'rxn_1', 'rxn_4', 'rxn_2' })
        self.assertEqual(self.model.matrix[Compound('A'), 'rxn_1'], -2)
        self.assertEqual(self.model.matrix[Compound('A'), 'rxn_2'], 1)
        self.assertEqual(self.model.matrix[Compound('B'), 'rxn_2'], -1)
        self.assertEqual(self.model.matrix[Compound('A'), 'rxn_4'], -1)
        self.assertEqual(self.model.matrix[Compound('C'), 'rxn_4'], 1)
        self.assertEqual(self.model.matrix[Compound('C'), 'rxn_5'], 1)
        self.assertEqual(self.model.matrix[Compound('D'), 'rxn_5'], -1)

    def test_flipable_model_view_limits_get_item_after_flip(self):
        self.model.flip({ 'rxn_1', 'rxn_2' })
        self.assertEqual(self.model.limits['rxn_1'].bounds, (-1000, 0))
        self.assertEqual(self.model.limits['rxn_2'].bounds, (-1000, 1000))
        self.assertEqual(self.model.limits['rxn_3'].bounds, (0, 1000))

    def test_flipable_model_view_limits_set_item_after_flip(self):
        self.model.flip({ 'rxn_1' })
        self.model.limits['rxn_1'].bounds = -20, 500
        self.assertEqual(self.model.limits['rxn_1'].bounds, (-20, 500))

        self.model.flip({ 'rxn_1' })
        self.assertEqual(self.model.limits['rxn_1'].bounds, (-500, 20))


if __name__ == '__main__':
    unittest.main()



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2015  Jon Lund Steffensen <jon_steffensen@uri.edu>

import logging

from ..command import Command, SolverCommandMixin
from ..gapfill import gapfind, gapfill

logger = logging.getLogger(__name__)


class GapFillCommand(SolverCommandMixin, Command):
    """Run the GapFind and GapFill algorithms on the model."""

    def run(self):
        """Run GapFill command"""

        # Load compound information
        compound_name = {}
        for compound in self._model.parse_compounds():
            if 'name' in compound.properties:
                compound_name[compound.id] = compound.properties['name']
            elif compound.id not in compound_name:
                compound_name[compound.id] = compound.id

        model_compartments = set(self._mm.compartments)
        core = set(self._mm.reactions)

        solver = self._get_solver(integer=True)

        # Run GapFind on model
        logger.info('Searching for blocked compounds')
        blocked = set(compound for compound in gapfind(self._mm, solver=solver)
                      if compound.compartment is not 'e')
        if len(blocked) > 0:
            logger.info('Blocked compounds')
            for compound in blocked:
                print(compound.translate(lambda x: compound_name.get(x, x)))

        if len(blocked) > 0:
            # Add exchange and transport reactions to database
            model_complete = self._mm.copy()
            logger.info('Adding database, exchange and transport reactions')
            model_complete.add_all_database_reactions(model_compartments)
            model_complete.add_all_exchange_reactions()
            model_complete.add_all_transport_reactions()

            logger.info('Searching for reactions to fill gaps')
            added_reactions, reversed_reactions = gapfill(
                model_complete, core, blocked, solver=solver)

            for rxnid in added_reactions:
                rx = model_complete.get_reaction(rxnid)
                rxt = rx.translated_compounds(
                    lambda x: compound_name.get(x, x))
                print('{}\t{}\t{}'.format(rxnid, 'Add', rxt))

            for rxnid in reversed_reactions:
                rx = model_complete.get_reaction(rxnid)
                rxt = rx.translated_compounds(
                    lambda x: compound_name.get(x, x))
                print('{}\t{}\t{}'.format(rxnid, 'Reverse', rxt))
        else:
            logger.info('No blocked compounds found')



#!/usr/bin/env python
# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2015  Jon Lund Steffensen <jon_steffensen@uri.edu>

import unittest

from psamm.metabolicmodel import MetabolicModel
from psamm.database import DictDatabase
from psamm import fluxcoupling
from psamm.datasource.modelseed import parse_reaction
from psamm.lpsolver import generic


class TestFluxCouplingBurgardModel(unittest.TestCase):
    """Test case based on the simple model in [Burgard04]_."""

    def setUp(self):
        self.database = DictDatabase()
        self.database.set_reaction('rxn_1', parse_reaction('|A| => |B|'))
        self.database.set_reaction('rxn_2', parse_reaction('|B| => |G|'))
        self.database.set_reaction('rxn_3', parse_reaction('|A| => |G|'))
        self.database.set_reaction('rxn_4', parse_reaction('|B| => |H|'))
        self.database.set_reaction('rxn_5', parse_reaction('|B| => |C| + |F|'))
        self.database.set_reaction('rxn_6', parse_reaction('|C| => |D|'))
        self.database.set_reaction('rxn_7', parse_reaction('|E| <=> |D|'))
        self.database.set_reaction('rxn_9', parse_reaction('|I| => |J|'))
        self.database.set_reaction('rxn_10', parse_reaction('|J| => |K|'))
        self.database.set_reaction('rxn_A', parse_reaction('=> |A|'))
        self.database.set_reaction('rxn_G', parse_reaction('|G| =>'))
        self.database.set_reaction('rxn_E', parse_reaction('|E| =>'))
        self.database.set_reaction(
            'rxn_bio', parse_reaction('|D| + (2.5) |F| =>'))
        self.model = MetabolicModel.load_model(
            self.database, self.database.reactions)

        try:
            self.solver = generic.Solver()
        except generic.RequirementsError:
            self.skipTest('Unable to find an LP solver for tests')

    def test_flux_coupling(self):
        if self.solver.properties['name'] == 'qsoptex':
            self.skipTest('Flux coupling with QSopt_ex is very slow')

        fcp = fluxcoupling.FluxCouplingProblem(self.model, {}, self.solver)
        reactions = sorted(self.model.reactions)
        couplings = {}
        for r1 in reactions:
            for r2 in reactions:
                couplings[r1, r2] = fcp.solve(r1, r2)

        inconsistent = {'rxn_4', 'rxn_9', 'rxn_10'}
        coupled_set = {'rxn_5', 'rxn_6', 'rxn_7', 'rxn_bio', 'rxn_E'}

        for i, r1 in enumerate(reactions):
            for r2 in reactions[i:]:
                if r1 == r2:
                    # A reaction is always fully coupled to itself
                    # unless it is inconsistent.
                    lower, upper = couplings[r1, r2]
                    if r1 in inconsistent:
                        self.assertEqual((lower, upper), (None, None))
                    else:
                        print('{}, {}: {}, {}'.format(r1, r2, lower, upper))
                        self.assertAlmostEqual(lower, upper)
                elif r1 in inconsistent or r2 in inconsistent:
                    # Couplings with inconsistent reactions are always
                    # fully coupled at 0.0 or uncoupled.
                    coupling = couplings[r1, r2]
                    self.assertIn(coupling, ((0.0, 0.0), (None, None)))
                elif r1 in coupled_set and r2 in coupled_set:
                    lower, upper = couplings[r1, r2]
                    self.assertAlmostEqual(lower, upper)
                else:
                    # At least one of the values must be 0 or None if the
                    # reactions are not in the coupled set.
                    coupling = couplings[r1, r2]
                    self.assertTrue(any(v in (0.0, None) for v in coupling))

        # Couplings with rxn_7 are negative
        for r in coupled_set:
            if r != 'rxn_7':
                self.assertLess(couplings['rxn_7', r][0], 0.0)


class TestFluxCouplingClass(unittest.TestCase):
    def test_uncoupled(self):
        cc = fluxcoupling.classify_coupling((None, None))
        self.assertEqual(cc, fluxcoupling.CouplingClass.Uncoupled)

    def test_directional_reverse(self):
        cc = fluxcoupling.classify_coupling((None, 1000))
        self.assertEqual(cc, fluxcoupling.CouplingClass.DirectionalReverse)

        cc = fluxcoupling.classify_coupling((-1000, None))
        self.assertEqual(cc, fluxcoupling.CouplingClass.DirectionalReverse)

    def test_inconsistent(self):
        cc = fluxcoupling.classify_coupling((0.0, 0.0))
        self.assertEqual(cc, fluxcoupling.CouplingClass.Inconsistent)

    def test_directional_forward(self):
        cc = fluxcoupling.classify_coupling((-1000, 100.0))
        self.assertEqual(cc, fluxcoupling.CouplingClass.DirectionalForward)

    def test_full(self):
        cc = fluxcoupling.classify_coupling((-1000, -1000.0))
        self.assertEqual(cc, fluxcoupling.CouplingClass.Full)


if __name__ == '__main__':
    unittest.main()



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2015  Jon Lund Steffensen <jon_steffensen@uri.edu>

"""Linear programming solver using Gurobi."""

from __future__ import absolute_import

import logging
from itertools import repeat, count
import numbers

from six.moves import zip

import gurobipy

from .lp import Solver as BaseSolver
from .lp import Constraint as BaseConstraint
from .lp import Problem as BaseProblem
from .lp import Result as BaseResult
from .lp import (Expression, Relation, ObjectiveSense, VariableType,
                 InvalidResultError)

# Module-level logging
logger = logging.getLogger(__name__)


class Solver(BaseSolver):
    """Represents an LP-solver using Gurobi."""

    def create_problem(self, **kwargs):
        """Create a new LP-problem using the solver."""
        return Problem(**kwargs)


class Problem(BaseProblem):
    """Represents an LP-problem of a gurobi.Solver."""

    VARTYPE_MAP = {
        VariableType.Continuous: gurobipy.GRB.CONTINUOUS,
        VariableType.Binary: gurobipy.GRB.BINARY,
        VariableType.Integer: gurobipy.GRB.INTEGER
    }

    CONSTR_SENSE_MAP = {
        Relation.Equals: gurobipy.GRB.EQUAL,
        Relation.Greater: gurobipy.GRB.GREATER_EQUAL,
        Relation.Less: gurobipy.GRB.LESS_EQUAL
    }

    OBJ_SENSE_MAP = {
        ObjectiveSense.Maximize: gurobipy.GRB.MAXIMIZE,
        ObjectiveSense.Minimize: gurobipy.GRB.MINIMIZE
    }

    def __init__(self, **kwargs):
        self._p = gurobipy.Model()

        # TODO Logging should use the Python logging system like every other
        # part of PSAMM. Unfortunately, Gurobi does not make this easy as the
        # logging output is directed to stdout/(stderr?) and/or a file. For
        # now, let's just disable logging to console so we don't mess up any
        # of our output. This can be done with the LogToConsole parameter which
        # disables output to console EXCEPT that when this parameter is set
        # a log message is generated explaining that the parameter was
        # changed... We are fortunate enough that this does not happen when
        # using the OutputFlag parameter but instead we lose the log file.
        self._p.params.OutputFlag = 0

        self._variables = {}
        self._var_names = ('x'+str(i) for i in count(1))
        self._constr_names = ('c'+str(i) for i in count(1))

        self._result = None

    @property
    def gurobi(self):
        """The underlying Gurobi Model object."""
        return self._p

    def define(self, *names, **kwargs):
        """Define variable in the problem.

        Variables must be defined before they can be accessed by var() or
        set(). This function takes keyword arguments lower and upper to define
        the bounds of the variable (default: -inf to inf). The keyword argument
        types can be used to select the type of the variable (Continuous
        (default), Binary or Integer). Setting any variables different than
        Continuous will turn the problem into an MILP problem.
        """
        names = tuple(names)
        lower = kwargs.get('lower', None)
        upper = kwargs.get('upper', None)
        vartype = kwargs.get('types', None)

        # Repeat values if a scalar is given
        if lower is None or isinstance(lower, numbers.Number):
            lower = repeat(lower, len(names))
        if upper is None or isinstance(upper, numbers.Number):
            upper = repeat(upper, len(names))
        if vartype is None or vartype in (
                VariableType.Continuous, VariableType.Binary,
                VariableType.Integer):
            vartype = repeat(vartype, len(names))

        lp_names = tuple(next(self._var_names) for name in names)

        # Assign default values
        lower = (-gurobipy.GRB.INFINITY if value is None else value
                 for value in lower)
        upper = (gurobipy.GRB.INFINITY if value is None else value
                 for value in upper)
        vartype = tuple(VariableType.Continuous if value is None else value
                        for value in vartype)

        self._variables.update(zip(names, lp_names))
        for name, lb, ub, vt in zip(lp_names, lower, upper, vartype):
            self._p.addVar(name=name, lb=lb, ub=ub, vtype=self.VARTYPE_MAP[vt])

        self._p.update()

    def has_variable(self, name):
        """Check whether variable is defined in the model."""
        return name in self._variables

    def _grb_expr_from_value_set(self, value_set):
        return gurobipy.LinExpr(
            [(float(val), self._p.getVarByName(self._variables[var]))
             for var, val in value_set])

    def _add_constraints(self, relation):
        """Add the given relation as one or more constraints.

        Return a list of the names of the constraints added.
        """
        names = []
        expression = relation.expression
        for value_set in expression.value_sets():
            name = next(self._constr_names)
            self._p.addConstr(
                self._grb_expr_from_value_set(value_set),
                self.CONSTR_SENSE_MAP[relation.sense],
                -float(expression.offset), name)
            names.append(name)

        self._p.update()

        return names

    def add_linear_constraints(self, *relations):
        """Add constraints to the problem.

        Each constraint is represented by a Relation, and the
        expression in that relation can be a set expression.
        """
        constraints = []

        for relation in relations:
            if self._check_relation(relation):
                constraints.append(Constraint(self, None))
            else:
                for name in self._add_constraints(relation):
                    constraints.append(Constraint(self, name))

        return constraints

    def set_linear_objective(self, expression):
        """Set linear objective of problem."""

        if isinstance(expression, numbers.Number):
            # Allow expressions with no variables as objective,
            # represented as a number
            expression = Expression()

        self._p.setObjective(
            self._grb_expr_from_value_set(expression.values()))

    def set_objective_sense(self, sense):
        """Set type of problem (maximize or minimize)."""

        if sense not in (ObjectiveSense.Minimize, ObjectiveSense.Maximize):
            raise ValueError('Invalid objective sense')

        self._p.ModelSense = self.OBJ_SENSE_MAP[sense]

    def solve(self, sense=None):
        """Solve problem."""
        if sense is not None:
            self.set_objective_sense(sense)
        self._p.optimize()

        self._result = Result(self)
        return self._result

    @property
    def result(self):
        return self._result


class Constraint(BaseConstraint):
    """Represents a constraint in a gurobi.Problem."""

    def __init__(self, prob, name):
        self._prob = prob
        self._name = name

    def delete(self):
        if self._name is not None:
            self._prob._p.remove(self._prob._p.getConstrByName(self._name))


class Result(BaseResult):
    """Represents the solution to a gurobi.Problem.

    This object will be returned from the gurobi.Problem.solve() method or by
    accessing the gurobi.Problem.result property after solving a problem. This
    class should not be instantiated manually.

    Result will evaluate to a boolean according to the success of the
    solution, so checking the truth value of the result will immediately
    indicate whether solving was successful.
    """

    def __init__(self, prob):
        self._problem = prob

    def _check_valid(self):
        if self._problem.result != self:
            raise InvalidResultError()

    @property
    def success(self):
        """Return boolean indicating whether a solution was found."""
        self._check_valid()
        return self._problem._p.Status == gurobipy.GRB.OPTIMAL

    @property
    def unbounded(self):
        """Whether solution is unbounded"""
        self._check_valid()

        status = self._problem._p.Status
        if (status == gurobipy.GRB.INF_OR_UNBD and
                self._problem._p.params.DualReductions):
            # Disable dual reductions to obtain a definitve answer
            self._problem._p.params.DualReductions = 0
            try:
                self._problem._p.optimize()
            finally:
                self._problem._p.params.DualReductions = 1

            status = self._problem._p.Status
        return status == gurobipy.GRB.UNBOUNDED

    @property
    def status(self):
        """Return string indicating the error encountered on failure."""
        self._check_valid()
        return str(self._problem._p.Status)

    def get_value(self, expression):
        """Return value of expression."""

        self._check_valid()
        if isinstance(expression, Expression):
            return sum(self._problem._p.getVarByName(
                self._problem._variables[var]).x * value
                for var, value in expression.values())
        elif expression not in self._problem._variables:
            raise ValueError('Unknown expression: {}'.format(expression))
        return self._problem._p.getVarByName(
            self._problem._variables[expression]).x



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2015  Jon Lund Steffensen <jon_steffensen@uri.edu>

"""Miscellaneaus data source functions."""

import re
from decimal import Decimal

from psamm.reaction import Reaction, Compound


def parse_sudensimple_reaction(s, arrow_rev='<=>', arrow_irrev='->'):
    """Parse a reaction string (SudenSimple)"""

    # Compile pattern for matching reaction arrows
    arrow_pattern = re.compile(
        '(' + ('|'.join(re.escape(x) for x in (arrow_irrev, arrow_rev))) + ')')

    def parse_compound_list(s):
        if s.strip() == '':
            return

        for cpd in s.strip().split('+'):
            cpd_split = cpd.strip().split(' ')
            if len(cpd_split) == 1:
                count = 1
                spec = cpd_split[0].strip()
            else:
                count = Decimal(cpd_split[0])
                if count % 1 == 0:
                    count = int(count)
                spec = cpd_split[1].strip()

            m = re.match(r'^(.+)\[(.+)\]$', spec)
            if m:
                # compartment
                cpdid = m.group(1)
                comp = m.group(2)
            else:
                cpdid = spec
                comp = None

            yield Compound(cpdid, compartment=comp), count

    # Split by equation arrow
    direction = Reaction.Right
    left, arrow, right = arrow_pattern.split(s, maxsplit=1)
    direction = Reaction.Right if arrow == arrow_irrev else Reaction.Bidir

    return Reaction(direction, parse_compound_list(left),
                    parse_compound_list(right))


def parse_metnet_reaction(s, arrow_rev='<==>', arrow_irrev='-->'):
    """Parser for the reaction format in MetNet model"""

    # Compile pattern for matching reaction arrows
    arrow_pattern = re.compile(
        '(' + ('|'.join(re.escape(x) for x in (arrow_irrev, arrow_rev))) + ')')

    def parse_compound_list(s, global_comp):
        if s.strip() == '':
            return

        for cpd in s.strip().split('+'):
            cpd_split = cpd.strip().split(') ')
            if len(cpd_split) == 1:
                count = 1
                spec = cpd_split[0].strip()
            else:
                count = Decimal(cpd_split[0].lstrip('('))
                if count % 1 == 0:
                    count = int(count)
                spec = cpd_split[1].strip()

            if global_comp is None:
                m = re.match(r'^(.+)\[(.+)\]$', spec)
                if m:
                    # compartment
                    cpdid = m.group(1)
                    comp = m.group(2)
                else:
                    cpdid = spec
                    comp = None
            else:
                cpdid = spec
                comp = global_comp

            yield Compound(cpdid, compartment=comp), count

    # Split by colon for compartment information
    m = re.match(r'^\s*\[(\w+)\]\s*:\s*(.*)', s)
    if m is not None:
        global_comp = m.group(1)
        s = m.group(2)
    else:
        global_comp = None

    # Split by equation arrow
    direction = Reaction.Right
    left, arrow, right = arrow_pattern.split(s, maxsplit=1)
    direction = Reaction.Right if arrow == arrow_irrev else Reaction.Bidir

    return Reaction(direction, parse_compound_list(left, global_comp),
                    parse_compound_list(right, global_comp))



# -*- coding: utf-8 -*-
# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2015  Jon Lund Steffensen <jon_steffensen@uri.edu>

"""Mass consistency analysis of metabolic databases

A stoichiometric matrix, S, is said to be mass-consistent if
S^Tm = 0 has a positive solution (m_i > 0). This corresponds to
assigning a positive mass to each compound in the stoichiometric
matrix and having each reaction preserve mass. Exchange reactions
will have to be excluded from this check, as they are not able to
preserve mass (by definition). In addition some databases may contain
pseudo-compounds (e.g. "photon") that also has to be excluded.
"""

from .lpsolver import lp

from six import iteritems


class MassConsistencyError(Exception):
    """Indicates an error while checking for mass consistency"""


def _non_localized_compounds(database):
    """Return set of non-localized compounds in database"""
    return set(c.in_compartment(None) for c in database.compounds)


def is_consistent(database, solver, exchange=set(), zeromass=set()):
    """Try to assign a positive mass to each compound

    Return True if successful. The masses are simply constrained by m_i > 1 and
    finding a solution under these conditions proves that the database is mass
    consistent.
    """

    prob = solver.create_problem()
    compound_set = _non_localized_compounds(database)

    # Define mass variables
    for compound in compound_set:
        if compound not in zeromass:
            prob.define(('m', compound), lower=1)
        else:
            prob.define(('m', compound), lower=0, upper=0)

    prob.set_linear_objective(prob.expr(
        {('m', compound): 1 for compound in compound_set}))

    # Define constraints
    massbalance_lhs = {reaction: 0 for reaction in database.reactions}
    for spec, value in iteritems(database.matrix):
        compound, reaction = spec
        mass = prob.var(('m', compound.in_compartment(None)))
        massbalance_lhs[reaction] += mass * value
    for reaction, lhs in iteritems(massbalance_lhs):
        if reaction not in exchange:
            prob.add_linear_constraints(lhs == 0)

    result = prob.solve(lp.ObjectiveSense.Minimize)
    return result.success


def check_reaction_consistency(database, solver, exchange=set(),
                               checked=set(), zeromass=set(), weights={}):
    """Check inconsistent reactions by minimizing mass residuals

    Return a reaction iterable, and compound iterable. The reaction iterable
    yields reaction ids and mass residuals. The compound iterable yields
    compound ids and mass assignments.

    Each compound is assigned a mass of at least one, and the masses are
    balanced using the stoichiometric matrix. In addition, each reaction has a
    residual mass that is included in the mass balance equations. The L1-norm
    of the residuals is minimized. Reactions in the checked set are assumed to
    have been manually checked and therefore have the residual fixed at zero.
    """

    # Create Flux balance problem
    prob = solver.create_problem()
    compound_set = _non_localized_compounds(database)

    # Define mass variables
    for compound in compound_set:
        if compound not in zeromass:
            prob.define(('m', compound), lower=1)
        else:
            prob.define(('m', compound), lower=0, upper=0)

    # Define residual mass variables and objective constriants
    prob.define(*(('z', reaction_id) for reaction_id in database.reactions),
                lower=0)
    prob.define(*(('r', reaction_id) for reaction_id in database.reactions))

    objective = prob.expr(
        {('z', reaction_id): weights.get(reaction_id, 1)
         for reaction_id in database.reactions})
    prob.set_linear_objective(objective)

    r = prob.set(('r', reaction_id) for reaction_id in database.reactions)
    z = prob.set(('z', reaction_id) for reaction_id in database.reactions)
    prob.add_linear_constraints(z >= r, r >= -z)

    massbalance_lhs = {reaction_id: 0 for reaction_id in database.reactions}
    for spec, value in iteritems(database.matrix):
        compound, reaction_id = spec
        mass_var = prob.var(('m', compound.in_compartment(None)))
        massbalance_lhs[reaction_id] += value * mass_var
    for reaction_id, lhs in iteritems(massbalance_lhs):
        if reaction_id not in exchange:
            if reaction_id not in checked:
                residual = prob.var(('r', reaction_id))
                prob.add_linear_constraints(lhs + residual == 0)
            else:
                prob.add_linear_constraints(lhs == 0)

    # Solve
    result = prob.solve(lp.ObjectiveSense.Minimize)
    if not result:
        raise MassConsistencyError('Non-optimal solution: {}'.format(
            result.status))

    def iterate_reactions():
        for reaction_id in database.reactions:
            residual = result.get_value(('r', reaction_id))
            yield reaction_id, residual

    def iterate_compounds():
        for compound in compound_set:
            yield compound, result.get_value(('m', compound))

    return iterate_reactions(), iterate_compounds()


def check_compound_consistency(database, solver, exchange=set(),
                               zeromass=set()):
    """Yield each compound in the database with assigned mass

    Each compound will be assigned a mass and the number of compounds having a
    positive mass will be approximately maximized.

    This is an implementation of the solution originally proposed by
    [Gevorgyan08]_  but using the new method proposed by [Thiele14]_ to avoid
    MILP constraints. This is similar to the way Fastcore avoids MILP
    contraints.
    """

    # Create mass balance problem
    prob = solver.create_problem()
    compound_set = _non_localized_compounds(database)

    # Define mass variables
    for compound in compound_set:
        if compound not in zeromass:
            prob.define(('m', compound), lower=0)
        else:
            prob.define(('m', compound), lower=0, upper=0)

    # Define z variables
    prob.define(*(('z', compound) for compound in compound_set),
                lower=0, upper=1)
    prob.set_linear_objective(prob.expr(
        {('z', compound): 1 for compound in compound_set}))

    z = prob.set(('z', compound) for compound in compound_set)
    m = prob.set(('m', compound) for compound in compound_set)
    prob.add_linear_constraints(m >= z)

    massbalance_lhs = {reaction_id: 0 for reaction_id in database.reactions}
    for spec, value in iteritems(database.matrix):
        compound, reaction_id = spec
        mass_var = prob.var(('m', compound.in_compartment(None)))
        massbalance_lhs[reaction_id] += value * mass_var
    for reaction_id, lhs in iteritems(massbalance_lhs):
        if reaction_id not in exchange:
            prob.add_linear_constraints(lhs == 0)

    # Solve
    result = prob.solve(lp.ObjectiveSense.Maximize)
    if not result:
        raise MassConsistencyError('Non-optimal solution: {}'.format(
            result.status))

    for compound in compound_set:
        yield compound, result.get_value(('m', compound))



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2015  Jon Lund Steffensen <jon_steffensen@uri.edu>

"""Parser for SBML model files"""

import xml.etree.ElementTree as ET
from decimal import Decimal
from fractions import Fraction
from functools import partial
from itertools import count
import logging

from six import itervalues, iteritems

from ..reaction import Reaction, Compound


logger = logging.getLogger(__name__)


# Level 1 namespaces
SBML_NS_L1 = 'http://www.sbml.org/sbml/level1'

# Level 2 namespaces
SBML_NS_L2 = 'http://www.sbml.org/sbml/level2'
SBML_NS_L2_V2 = 'http://www.sbml.org/sbml/level2/version2'
SBML_NS_L2_V3 = 'http://www.sbml.org/sbml/level2/version3'
SBML_NS_L2_V4 = 'http://www.sbml.org/sbml/level2/version4'
SBML_NS_L2_V5 = 'http://www.sbml.org/sbml/level2/version5'

# Level 3 namespaces
SBML_NS_L3_V1_CORE = 'http://www.sbml.org/sbml/level3/version1/core'

MATHML_NS = 'http://www.w3.org/1998/Math/MathML'

# FBC namespaces
FBC_V1 = 'http://www.sbml.org/sbml/level3/version1/fbc/version1'
FBC_V2 = 'http://www.sbml.org/sbml/level3/version1/fbc/version2'


def _tag(tag, namespace=None):
    """Prepend namespace to tag name"""
    if namespace is None:
        return str(tag)
    return '{{{}}}{}'.format(namespace, tag)


class ParseError(Exception):
    """Error parsing SBML file"""


class _SBMLEntry(object):
    """Base class for compound and reaction entries"""

    def __init__(self, reader, root):
        self._reader = reader
        self._root = root
        self._id = self._element_get_id(root)

    def _element_get_id(self, element):
        """Get id of reaction or species element.

        In old levels the name is used as the id. This method returns the
        correct attribute depending on the level.
        """
        if self._reader._level > 1:
            entry_id = element.get('id')
        else:
            entry_id = element.get('name')
        return entry_id

    @property
    def id(self):
        """Entity ID"""
        return self._id

    @property
    def xml_notes(self):
        """Access the entity notes as an XML document fragment"""
        return self._root.find(self._reader._sbml_tag('notes'))


class SpeciesEntry(_SBMLEntry):
    """Species entry in the SBML file"""

    def __init__(self, reader, root):
        super(SpeciesEntry, self).__init__(reader, root)

        self._name = root.get('name')
        self._comp = root.get('compartment')

        self._boundary = root.get('boundaryCondition', 'false') == 'true'

        # In non-strict mode the species that ends with _b are considered
        # boundary conditions.
        if not self._reader._strict and self._id.endswith('_b'):
            logger.warning('Species {} was converted to boundary condition'
                           ' because of "_b" suffix'.format(self.id))
            self._boundary = True

    @property
    def name(self):
        """Species name"""
        return self._name

    @property
    def compartment(self):
        """Species compartment"""
        return self._comp

    def _parse_charge_string(self, s):
        try:
            return int(s)
        except ValueError:
            if self._reader._strict:
                raise ParseError('Invalid charge for species {}'.format(
                    self.id))
            else:
                return None

    @property
    def charge(self):
        """Species charge"""
        if self._reader._level == 3:
            # Look for FBC charge
            for ns in (FBC_V2, FBC_V1):
                charge = self._root.get(_tag('charge', ns))
                if charge is not None:
                    return self._parse_charge_string(charge)
        else:
            charge = self._root.get('charge')
            if charge is not None:
                return self._parse_charge_string(charge)

        return None

    @property
    def formula(self):
        """Species formula"""
        if self._reader._level == 3:
            for ns in (FBC_V2, FBC_V1):
                formula = self._root.get(_tag('chemicalFormula', ns))
                if formula is not None:
                    return formula

        return None

    @property
    def boundary(self):
        """Whether this compound is a boundary condition"""
        return self._boundary

    @property
    def properties(self):
        """All species properties as a dict"""
        properties = {'id': self._id,
                      'boundary': self._boundary}
        if 'name' in self._root.attrib:
            properties['name'] = self._root.get('name')
        if 'compartment' in self._root.attrib:
            properties['compartment'] = self._root.get('compartment')
        if 'charge' in self._root.attrib and self._root.get('charge') != '':
            properties['charge'] = int(self._root.get('charge'))

        return properties


class ReactionEntry(_SBMLEntry):
    """Reaction entry in SBML file"""

    def __init__(self, reader, root):
        super(ReactionEntry, self).__init__(reader, root)

        self._name = self._root.get('name')
        self._rev = self._root.get('reversible', 'true') == 'true'

        left, right = [], []
        for side, tag_name in ((left, 'listOfReactants'),
                               (right, 'listOfProducts')):
            for species_id, value in self._parse_species_references(tag_name):
                try:
                    species_entry = self._reader.get_species(species_id)
                    if (self._reader._ignore_boundary and
                            species_entry.boundary):
                        # Skip boundary species when ignoring these
                        continue
                except KeyError:
                    message = ('Reaction {} references non-existent'
                               ' species {}'.format(self._id, species_id))
                    if not self._reader._strict:
                        # In non-strict mode simply skip these references
                        logger.warn(message)
                        continue
                    raise ParseError(message)

                species_id, species_comp = (
                    species_entry.id, species_entry.compartment)
                compound = Compound(species_id, compartment=species_comp)
                side.append((compound, value))

        direction = Reaction.Bidir if self._rev else Reaction.Right
        self._equation = Reaction(direction, left, right)

        # Parse flux bounds of reaction
        self._lower_flux = None
        self._upper_flux = None

        if reader._level == 3:
            lower = self._root.get(_tag('lowerFluxBound', FBC_V2))
            if lower is not None:
                if lower not in reader._model_constants:
                    raise ParseError(
                        'Lower flux parameter of {} is not defined in'
                        ' model parameter constants'.format(self.id))
                self._lower_flux = reader._model_constants[lower]

            upper = self._root.get(_tag('upperFluxBound', FBC_V2))
            if upper is not None:
                if upper not in reader._model_constants:
                    raise ParseError(
                        'Upper flux parameter of {} is not defined in'
                        ' model parameter constants'.format(self.id))
                self._upper_flux = reader._model_constants[upper]

            if (lower is None and upper is None and
                    self.id in reader._reaction_flux_bounds):
                # Parse bounds from listOfFluxBounds in FBCv1
                for bound in reader._reaction_flux_bounds[self.id]:
                    if bound.operation == 'equal':
                        self._lower_flux = bound.value
                        self._upper_flux = bound.value
                    elif bound.operation == 'lessEqual':
                        self._upper_flux = bound.value
                    elif bound.operation == 'greaterEqual':
                        self._lower_flux = bound.value

    def _parse_species_references(self, name):
        """Yield species id and parsed value for a speciesReference list"""
        for species in self._root.iterfind('./{}/{}'.format(
                self._reader._sbml_tag(name),
                self._reader._sbml_tag('speciesReference'))):

            species_id = species.get('species')

            if self._reader._level == 1:
                # In SBML level 1 only positive integers are allowed for
                # stoichiometry but a positive integer denominator can be
                # specified.
                try:
                    value = int(species.get('stoichiometry', 1))
                    denom = int(species.get('denominator', 1))
                    species_value = Fraction(value, denom)
                except ValueError:
                    message = ('Non-integer stoichiometry is not allowed in'
                               ' SBML level 1 (reaction {})'.format(self.id))
                    if self._reader._strict:
                        raise ParseError(message)
                    else:
                        logger.warning(message)
                    species_value = Decimal(species.get('stoichiometry', 1))
            elif self._reader._level == 2:
                # Stoichiometric value is a double but can alternatively be
                # specified using math (not implemented).
                value_str = species.get('stoichiometry', None)
                if value_str is None:
                    if 'stoichiometryMath' in species:
                        raise ParseError('stoichiometryMath in '
                                         'speciesReference is not implemented')
                    species_value = 1
                else:
                    species_value = Decimal(value_str)
            elif self._reader._level == 3:
                # Stoichiometric value is a double but can alternatively be
                # specified using initial assignment (not implemented).
                value_str = species.get('stoichiometry', None)
                if value_str is None:
                    raise ParseError('Missing stoichiometry in'
                                     ' speciesReference is not implemented')
                species_value = Decimal(value_str)

            if species_value % 1 == 0:
                species_value = int(species_value)

            yield species_id, species_value

    @property
    def id(self):
        """Reaction ID"""
        return self._id

    @property
    def name(self):
        """Reaction name"""
        return self._name

    @property
    def reversible(self):
        """Whether the reaction is reversible"""
        return self._rev

    @property
    def equation(self):
        """Reaction equation is a :class:`Reaction <psamm.reaction.Reaction>`
        object"""
        return self._equation

    @property
    def kinetic_law_reaction_parameters(self):
        """Iterator over the values of kinetic law reaction parameters"""

        for parameter in self._root.iterfind(
                './{}/{}/{}'.format(self._reader._sbml_tag('kineticLaw'),
                                    self._reader._sbml_tag('listOfParameters'),
                                    self._reader._sbml_tag('parameter'))):
            param_id = parameter.get('id')
            param_name = parameter.get('name')
            param_value = float(parameter.get('value'))
            param_units = parameter.get('units')

            yield param_id, param_name, param_value, param_units

    @property
    def properties(self):
        """All reaction properties as a dict"""
        properties = {'id': self._id,
                      'reversible': self._rev,
                      'equation': self._equation}
        if 'name' in self._root.attrib:
            properties['name'] = self._root.get('name')
        if self._lower_flux is not None:
            properties['lower_flux'] = self._lower_flux
        if self._upper_flux is not None:
            properties['upper_flux'] = self._upper_flux

        return properties


class ObjectiveEntry(object):
    """Flux objective defined with FBC"""

    def __init__(self, reader, namespace, root):
        self._reader = reader
        self._root = root
        self._namespace = namespace

        self._id = self._root.get(_tag('id', self._namespace))
        if self._id is None:
            raise ParseError('Objective has no "id" attribute')

        self._name = self._root.get(_tag('name', self._namespace))
        self._type = self._root.get(_tag('type', self._namespace))

        if self._type is None:
            raise ParseError('Missing "type" attribute on objective: {}'.formt(
                self.id))

        # Find flux objectives
        self._reactions = {}
        for fo in self._root.iterfind('./{}/{}'.format(
                _tag('listOfFluxObjectives', self._namespace),
                _tag('fluxObjective', self._namespace))):
            reaction = fo.get(_tag('reaction', self._namespace))
            coefficient = float(fo.get(_tag('coefficient', self._namespace)))
            if reaction is None or coefficient is None:
                raise ParseError(
                    'Missing attributes on flux objective: {}'.format(self.id))

            if coefficient != 0:
                self._reactions[reaction] = coefficient

    @property
    def id(self):
        return self._id

    @property
    def name(self):
        return self._name

    @property
    def type(self):
        return self._type

    @property
    def reactions(self):
        return iteritems(self._reactions)


class FluxBoundEntry(object):
    """Flux bound defined with FBC"""

    def __init__(self, reader, namespace, root):
        self._reader = reader
        self._root = root
        self._namespace = namespace

        self._id = self._root.get(_tag('id', self._namespace))
        self._name = self._root.get(_tag('name', self._namespace))

        self._reaction = self._root.get(_tag('reaction', self._namespace))
        if self._reaction is None:
            raise ParseError('Flux bound is missing "reaction" attribute')

        self._operation = self._root.get(_tag('operation', self._namespace))
        if self._operation is None:
            raise ParseError('Flux bound is missing "operation" attribute')
        elif self._operation not in ('lessEqual', 'greaterEqual', 'equal'):
            raise ParseError('Invalid flux bound operation: {}'.format(
                self._operation))

        value = self._root.get(_tag('value', self._namespace))
        if value is None:
            raise ParseError('Flux bound is missing "value" attribute')
        try:
            self._value = float(value)
        except ValueError:
            raise ParseError('Unable to parse flux bound value')

    @property
    def id(self):
        return self._id

    @property
    def name(self):
        return self._name

    @property
    def reaction(self):
        return self._reaction

    @property
    def operation(self):
        return self._operation

    @property
    def value(self):
        return self._value


class SBMLReader(object):
    """Reader of SBML model files

    The constructor takes a file-like object which will be parsed as XML and
    then as SBML according to the specification. If the ``strict`` parameter is
    set to False, the parser will revert to a more lenient parsing which is
    required for many older models. This tries to mimic the inconsistencies
    employed by COBRA when parsing models.

    If ``ignore_boundary`` is ``True``, the species that are marked as
    boundary conditions will simply be dropped from the species list and from
    the reaction equations.
    """

    def __init__(self, file, strict=False, ignore_boundary=False):
        # Parse SBML file
        tree = ET.parse(file)
        root = tree.getroot()

        self._strict = strict
        self._ignore_boundary = ignore_boundary

        # Parse level and version
        self._sbml_tag = None
        self._level = int(root.get('level'))
        self._version = int(root.get('version'))

        if self._level == 1:
            self._sbml_tag = partial(_tag, namespace=SBML_NS_L1)
        elif self._level == 2:
            if self._version == 1:
                self._sbml_tag = partial(_tag, namespace=SBML_NS_L2)
            elif self._version == 2:
                self._sbml_tag = partial(_tag, namespace=SBML_NS_L2_V2)
            elif self._version == 3:
                self._sbml_tag = partial(_tag, namespace=SBML_NS_L2_V3)
            elif self._version == 4:
                self._sbml_tag = partial(_tag, namespace=SBML_NS_L2_V4)
            elif self._version == 5:
                self._sbml_tag = partial(_tag, namespace=SBML_NS_L2_V5)
            else:
                if self._strict:
                    raise ParseError('SBML level 2, version {}'
                                     ' not implemented'.format(self._version))
                # Assume latest version
                self._version = 5
                self._sbml_tag = partial(_tag, namespace=SBML_NS_L2_V5)
        elif self._level == 3:
            if self._version == 1:
                self._sbml_tag = partial(_tag, namespace=SBML_NS_L3_V1_CORE)
            else:
                if self._strict:
                    raise ParseError('SBML level 3, version {}'
                                     ' not implemented'.format(self._version))
                # Assume latest version
                self._version = 1
                self._sbml_tag = partial(_tag, namespace=SBML_NS_L3_V1_CORE)
        else:
            if self._strict:
                raise ParseError('SBML level {} not implemented'.format(
                    self._level))
            # Assume level 1
            self._level = 1
            self._sbml_tag = partial(_tag, namespace=SBML_NS_L1)

        self._model = root.find(self._sbml_tag('model'))

        # Parameters
        self._model_constants = {}
        params = self._model.find(self._sbml_tag('listOfParameters'))
        if params is not None:
            for param in params.iterfind(self._sbml_tag('parameter')):
                if param.get('constant') == 'true':
                    param_id = param.get('id')
                    value = float(param.get('value'))
                    self._model_constants[param_id] = value

        # Flux bounds
        self._flux_bounds = []
        self._reaction_flux_bounds = {}
        if self._level == 3:
            # Only suported in FBC V1
            flux_bounds = self._model.find(_tag('listOfFluxBounds', FBC_V1))
            if flux_bounds is not None:
                for flux_bound in flux_bounds.iterfind(
                        _tag('fluxBound', FBC_V1)):
                    entry = FluxBoundEntry(self, FBC_V1, flux_bound)
                    self._flux_bounds.append(entry)

                    # Create reference from reaction to flux bound
                    entries = self._reaction_flux_bounds.setdefault(
                        entry.reaction, [])
                    entries.append(entry)

        # Species
        self._model_species = {}
        self._species = self._model.find(self._sbml_tag('listOfSpecies'))
        for species in self._species.iterfind(self._sbml_tag('species')):
            entry = SpeciesEntry(self, species)
            self._model_species[entry.id] = entry

        # Reactions
        self._model_reactions = {}
        self._reactions = self._model.find(self._sbml_tag('listOfReactions'))
        for reaction in self._reactions.iterfind(self._sbml_tag('reaction')):
            entry = ReactionEntry(self, reaction)
            self._model_reactions[entry.id] = entry

        # Objectives
        self._model_objectives = {}
        self._active_objective = None
        if self._level == 3:
            objectives = None
            for fbc_ns in (FBC_V2, FBC_V1):
                objectives = self._model.find(_tag('listOfObjectives', fbc_ns))
                if objectives is not None:
                    break

            if objectives is not None:
                for objective in objectives.iterfind(
                        _tag('objective', fbc_ns)):
                    entry = ObjectiveEntry(self, fbc_ns, objective)
                    self._model_objectives[entry.id] = entry

                active = objectives.get(_tag('activeObjective', fbc_ns))
                if active is None or active not in self._model_objectives:
                    raise ParseError('Active objective is invalid: {}'.format(
                        active))

                self._active_objective = self._model_objectives[active]

    def get_reaction(self, reaction_id):
        """Return :class:`.ReactionEntry` corresponding to reaction_id"""
        return self._model_reactions[reaction_id]

    def get_species(self, species_id):
        """Return :class:`.SpeciesEntry` corresponding to species_id"""
        return self._model_species[species_id]

    def get_objective(self, objective_id):
        """Return :class:`.ObjectiveEntry` corresponding to objective_id"""
        return self._model_objectives[objective_id]

    def get_active_objective(self):
        return self._active_objective

    @property
    def reactions(self):
        """Iterator over :class:`ReactionEntries <.ReactionEntry>`"""
        return itervalues(self._model_reactions)

    @property
    def species(self):
        """Iterator over :class:`SpeciesEntries <.SpeciesEntry>`

        This will not yield boundary condition species if those are ignored.
        """
        return (c for c in itervalues(self._model_species)
                if not self._ignore_boundary or not c.boundary)

    @property
    def objectives(self):
        """Iterator over :class:`.ObjectiveEntry`"""
        return itervalues(self._model_objectives)

    @property
    def flux_bounds(self):
        """Iterator over :class:`.FluxBoundEntry`"""
        return iter(self._flux_bounds)

    @property
    def id(self):
        """Model ID"""
        return self._model.get('id', None)

    @property
    def name(self):
        """Model name"""
        return self._model.get('name', None)


class SBMLWriter(object):
    """Writer of SBML files"""

    def __init__(self):
        self._namespace = SBML_NS_L3_V1_CORE
        self._sbml_tag = partial(_tag, namespace=self._namespace)

    def write_model(self, file, model, compounds):
        """Write a given model to file"""

        ET.register_namespace('mathml', MATHML_NS)

        # Load compound information
        compound_name = {}
        for compound in compounds:
            compound_name[compound.id] = (
                compound.name if compound.name is not None else compound.id)

        root = ET.Element(self._sbml_tag('sbml'))
        root.set(self._sbml_tag('level'), '3')
        root.set(self._sbml_tag('version'), '1')

        model_tag = ET.SubElement(root, self._sbml_tag('model'))

        # Generators of unique IDs
        compound_id = ('M_'+str(i) for i in count(1))
        compartment_id = ('C_'+str(i) for i in count(1))
        reaction_id = ('R_'+str(i) for i in count(1))

        # Build mapping from Compound to species ID
        model_compartments = {}
        model_species = {}
        for reaction in model.reactions:
            for compound, value in model.get_reaction_values(reaction):
                if compound.compartment not in model_compartments:
                    model_compartments[compound.compartment] = (
                        next(compartment_id))
                if compound not in model_species:
                    model_species[compound] = next(compound_id)

        # Create list of compartments
        compartments = ET.SubElement(
            model_tag, self._sbml_tag('listOfCompartments'))
        for compartment, compartment_id in iteritems(model_compartments):
            compartment_tag = ET.SubElement(
                compartments, self._sbml_tag('compartment'))
            compartment_tag.set(self._sbml_tag('id'), compartment_id)
            compartment_tag.set(self._sbml_tag('name'), str(compartment))

        # Create list of species
        species_list = ET.SubElement(
            model_tag, self._sbml_tag('listOfSpecies'))
        for species, species_id in iteritems(model_species):
            species_tag = ET.SubElement(species_list,
                                        self._sbml_tag('species'))
            species_tag.set(self._sbml_tag('id'), species_id)
            species_tag.set(
                self._sbml_tag('name'),
                str(species.translate(lambda x: compound_name.get(x, x))))
            species_tag.set(
                self._sbml_tag('compartment'),
                model_compartments[species.compartment])

        # Create list of reactions
        reactions = ET.SubElement(model_tag, self._sbml_tag('listOfReactions'))
        for reaction in model.reactions:
            reaction_tag = ET.SubElement(reactions, self._sbml_tag('reaction'))
            reaction_tag.set(self._sbml_tag('id'), next(reaction_id))
            reaction_tag.set(self._sbml_tag('name'), reaction)
            reaction_tag.set(
                self._sbml_tag('reversible'),
                'true' if model.is_reversible(reaction) else 'false')

            reactants = ET.SubElement(
                reaction_tag, self._sbml_tag('listOfReactants'))
            products = ET.SubElement(
                reaction_tag, self._sbml_tag('listOfProducts'))

            for compound, value in model.get_reaction_values(reaction):
                dest_list = reactants if value < 0 else products
                spec_ref = ET.SubElement(
                    dest_list, self._sbml_tag('speciesReference'))
                spec_ref.set(
                    self._sbml_tag('species'), model_species[compound])
                spec_ref.set(self._sbml_tag('stoichiometry'), str(abs(value)))

            # Create COBRA-compliant parameter list
            kl_tag = ET.SubElement(reaction_tag, self._sbml_tag('kineticLaw'))
            math_tag = ET.SubElement(kl_tag, self._sbml_tag('math'))
            ci_tag = ET.SubElement(math_tag, _tag('ci', MATHML_NS))
            ci_tag.text = 'FLUX_VALUE'
            param_list = ET.SubElement(
                kl_tag, self._sbml_tag('listOfParameters'))
            ET.SubElement(param_list, self._sbml_tag('parameter'), {
                self._sbml_tag('id'): 'LOWER_BOUND',
                self._sbml_tag('name'): 'LOWER_BOUND',
                self._sbml_tag('value'): str(model.limits[reaction].lower)
            })
            ET.SubElement(param_list, self._sbml_tag('parameter'), {
                self._sbml_tag('id'): 'UPPER_BOUND',
                self._sbml_tag('name'): 'UPPER_BOUND',
                self._sbml_tag('value'): str(model.limits[reaction].upper)
            })

        tree = ET.ElementTree(root)
        tree.write(file, default_namespace=self._namespace)



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2015  Jon Lund Steffensen <jon_steffensen@uri.edu>

"""Utilities for keeping track of parsing context."""


import os

from six import string_types


class FilePathContext(object):
    """File context that keeps track of contextual information.

    When a file is loaded, all files specified in that file must be loaded
    relative to the first file. This is made possible by keeping a context
    that remembers where a file was loaded so that other files can be loaded
    relatively.
    """

    def __init__(self, arg):
        """Create new context from a path or existing context"""

        if isinstance(arg, string_types):
            self._filepath = arg
        else:
            self._filepath = arg.filepath
        self._basepath = os.path.dirname(self._filepath)

    @property
    def filepath(self):
        return self._filepath

    @property
    def basepath(self):
        return self._basepath

    def resolve(self, relpath):
        return FilePathContext(os.path.join(self._basepath, relpath))

    def open(self, mode='r'):
        return open(self.filepath, mode)

    def __str__(self):
        return self._filepath


class FileMark(object):
    """Marks a position in a file.

    This is used when parsing input files, to keep track of the position that
    generates an entry.
    """

    def __init__(self, filecontext, line, column):
        self._filecontext = filecontext
        self._line = line
        self._column = column

    @property
    def filecontext(self):
        return self._filecontext

    @property
    def line(self):
        return self._line

    @property
    def column(self):
        return self._column

    def __str__(self):
        return '{}:{}:{}'.format(
            str(self._filecontext), '?' if self._line is None else self._line,
            '?' if self._column is None else self._column)

    def __repr__(self):
        return '{}({}, {}, {})'.format(
            self.__class__.__name__, repr(self._filecontext), repr(self._line),
            repr(self._column))



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2015  Jon Lund Steffensen <jon_steffensen@uri.edu>

import sys
import os
import shutil
import tempfile
from contextlib import contextmanager
import unittest

from six import StringIO, BytesIO

from psamm.command import main, Command, SolverCommandMixin
from psamm.lpsolver import generic


@contextmanager
def redirected_stdout(target=None):
    stdout = sys.stdout
    if target is None:
        target = StringIO()
    try:
        sys.stdout = target
        yield target
    finally:
        sys.stdout = stdout


class MockCommand(Command):
    """Test command.

    This is a command for testing basic commands.
    """
    init_parser_called = False
    run_called = False
    has_native_model = False
    has_metabolic_model = False

    @classmethod
    def init_parser(cls, parser):
        cls.init_parser_called = True
        parser.add_argument('--test-argument', action='store_true')
        super(MockCommand, cls).init_parser(parser)

    def __init__(self, *args, **kwargs):
        super(MockCommand, self).__init__(*args, **kwargs)

    def run(self):
        self.__class__.run_called = True
        self.__class__.has_native_model = hasattr(self, '_model')
        self.__class__.has_metabolic_model = hasattr(self, '_mm')


class MockSolverCommand(SolverCommandMixin, Command):
    """Test solver command.

    This is a command for testing solver commands.
    """
    def run(self):
        solver = self._get_solver()


class TestCommandMain(unittest.TestCase):
    def setUp(self):
        self._model_dir = tempfile.mkdtemp()
        with open(os.path.join(self._model_dir, 'model.yaml'), 'w') as f:
            f.write('''---
                name: Test model
                biomass: rxn_1
                reactions:
                  - id: rxn_1
                    equation: '|A[e]| => |B[c]|'
                    genes:
                      - gene_1
                      - gene_2
                  - id: rxn_2
                    equation: '|B[c]| => |C[e]|'
                    genes: 'gene_3 or (gene_4 and gene_5)'
                compounds:
                  - id: A
                  - id: B
                  - id: C
                media:
                  - compartment: e
                    compounds:
                      - id: A
                      - id: C
                limits:
                  - reaction: rxn_2
                    upper: 100
                ''')

    def tearDown(self):
        shutil.rmtree(self._model_dir)

    def is_solver_available(self, **kwargs):
        try:
            generic.Solver(**kwargs)
        except generic.RequirementsError:
            return False

        return True

    def run_solver_command(self, args, requirements):
        with redirected_stdout() as f:
            if self.is_solver_available(**requirements):
                main(args=args)
            else:
                with self.assertRaises(generic.RequirementsError):
                    main(args=args)

        return f

    def test_invoke_version(self):
        with redirected_stdout():
            with self.assertRaises(SystemExit) as context:
                main(args=['--version'])
            self.assertEqual(context.exception.code, 0)

    def test_run_chargecheck(self):
        with redirected_stdout():
            main(args=['--model', self._model_dir, 'chargecheck'])

    def test_run_excelexport(self):
        dest_path = os.path.join(self._model_dir, 'model.xlsx')
        with redirected_stdout():
            main(args=['--model', self._model_dir, 'excelexport', dest_path])

        self.assertTrue(os.path.isfile(dest_path))

    def test_run_fastgapfill(self):
        self.run_solver_command([
            '--model', self._model_dir, 'fastgapfill'], {})

    def test_run_fba(self):
        self.run_solver_command(['--model', self._model_dir, 'fba'], {})

    def test_run_fba_with_tfba(self):
        self.run_solver_command([
            '--model', self._model_dir, 'fba', '--loop-removal', 'tfba'],
            {'integer': True})

    def test_run_fba_with_l1min(self):
        self.run_solver_command([
            '--model', self._model_dir, 'fba', '--loop-removal', 'l1min'], {})

    def test_run_fluxcheck(self):
        self.run_solver_command([
            '--model', self._model_dir, 'fluxcheck'], {})

    def test_run_fluxcheck_with_fastcore(self):
        self.run_solver_command([
            '--model', self._model_dir, 'fluxcheck', '--fastcore'], {})

    def test_run_fluxcheck_with_tfba(self):
        self.run_solver_command([
            '--model', self._model_dir, 'fluxcheck', '--tfba'],
            {'integer': True})

    def test_run_fluxcheck_with_reduce_lp(self):
        self.run_solver_command([
            '--model', self._model_dir, 'fluxcheck', '--reduce-lp'], {})

    def test_run_fluxcheck_with_both_tfba_and_fastcore(self):
        with self.assertRaises(SystemExit):
            self.run_solver_command([
                '--model', self._model_dir, 'fluxcheck',
                '--tfba', '--fastcore'], {})

    def test_run_fluxcoupling(self):
        self.run_solver_command([
            '--model', self._model_dir, 'fluxcoupling'], {})

    def test_run_formulacheck(self):
        with redirected_stdout():
            main(args=['--model', self._model_dir, 'formulacheck'])

    def test_run_fva(self):
        self.run_solver_command([
            '--model', self._model_dir, 'fva'], {})

    def test_run_gapfill(self):
        self.run_solver_command([
            '--model', self._model_dir, 'gapfill'], {'integer': True})

    def test_run_masscheck_compounds(self):
        self.run_solver_command([
            '--model', self._model_dir, 'masscheck', '--type', 'compound'], {})

    def test_run_masscheck_reactions(self):
        self.run_solver_command([
            '--model', self._model_dir, 'masscheck', '--type', 'reaction'], {})

    def test_run_randomsparse_reactions(self):
        self.run_solver_command([
            '--model', self._model_dir, 'randomsparse', '--type=reactions',
            '50%'], {})

    def test_run_randomsparse_genes(self):
        self.run_solver_command([
            '--model', self._model_dir, 'randomsparse', '--type=genes', '50%'],
            {})

    def test_run_randomsparse_exchange(self):
        self.run_solver_command([
            '--model', self._model_dir, 'randomsparse', '--type=exchange',
            '50%'], {})

    def test_run_robustness(self):
        self.run_solver_command([
            '--model', self._model_dir, 'robustness', 'rxn_2'], {})

    def test_run_sbmlexport(self):
        with redirected_stdout(BytesIO()):
            main(args=['--model', self._model_dir, 'sbmlexport'])

    def test_run_search_compound(self):
        with redirected_stdout():
            main(args=[
                '--model', self._model_dir, 'search', 'compound', '--id', 'A'])

    def test_run_search_reaction(self):
        with redirected_stdout():
            main(args=[
                '--model', self._model_dir, 'search', 'reaction',
                '--id', 'rxn_1'])

    def test_run_tableexport_reactions(self):
        with redirected_stdout() as f:
            main(args=[
                '--model', self._model_dir, 'tableexport', 'reactions'])

        self.assertEqual(f.getvalue(), '\n'.join([
            'id\tequation\tgenes',
            "rxn_1\t|A[e]| => |B[c]|\t['gene_1', 'gene_2']",
            'rxn_2\t|B[c]| => |C[e]|\tgene_3 or (gene_4 and gene_5)',
            ''
        ]))

    def test_run_tableexport_compounds(self):
        with redirected_stdout() as f:
            main(args=[
                '--model', self._model_dir, 'tableexport', 'compounds'])

        self.assertEqual(f.getvalue(), '\n'.join([
            'id', 'A', 'B', 'C', ''
        ]))

    def test_run_tableexport_medium(self):
        with redirected_stdout() as f:
            main(args=[
                '--model', self._model_dir, 'tableexport', 'medium'])

        self.assertEqual(f.getvalue(), '\n'.join([
            'Compound ID\tReaction ID\tLower Limit\tUpper Limit',
            'A[e]\tNone\t-1000\t1000',
            'C[e]\tNone\t-1000\t1000',
            ''
        ]))

    def test_run_tableexport_limits(self):
        with redirected_stdout() as f:
            main(args=[
                '--model', self._model_dir, 'tableexport', 'limits'])

        self.assertEqual(f.getvalue(), '\n'.join([
            'Reaction ID\tLower Limits\tUpper Limits',
            'rxn_2\tNone\t100',
            ''
        ]))

    def test_run_tableexport_metadata(self):
        with redirected_stdout() as f:
            main(args=[
                '--model', self._model_dir, 'tableexport', 'metadata'])

        self.assertEqual(f.getvalue(), '\n'.join([
            'Model Name: Test model',
            'Biomass Reaction: rxn_1',
            'Default Flux Limits: None',
            ''
        ]))

    def test_command_main(self):
        main(MockCommand, args=[
            '--model', self._model_dir, '--test-argument'])
        self.assertTrue(MockCommand.run_called)
        self.assertTrue(MockCommand.init_parser_called)
        self.assertTrue(MockCommand.has_native_model)
        self.assertTrue(MockCommand.has_metabolic_model)

    def test_solver_command_main(self):
        with self.assertRaises(generic.RequirementsError):
            main(MockSolverCommand, args=[
                '--model', self._model_dir,
                '--solver', 'name=not-an-actual-solver'])



#!/usr/bin/env python
# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2015  Jon Lund Steffensen <jon_steffensen@uri.edu>

import unittest

from psamm.reaction import Reaction, Compound


class TestCompound(unittest.TestCase):
    def test_compound_init_no_arguments(self):
        with self.assertRaises(TypeError):
            c = Compound()

    def test_compound_init_with_name(self):
        c = Compound('Phosphate')

    def test_compound_init_with_compartment(self):
        c = Compound('Phosphate', compartment='e')

    def test_compound_init_with_argument(self):
        c = Compound('Menaquinone', arguments=[8])

    def test_compound_init_with_compartment_and_argument(self):
        c = Compound('Menaquinone', compartment='e', arguments=[8])

    def test_compound_name(self):
        c = Compound('Phosphate')
        self.assertEqual(c.name, 'Phosphate')

    def test_compound_compartment(self):
        c = Compound('Phosphate', compartment='e')
        self.assertEqual(c.compartment, 'e')

    def test_compound_compartment_none(self):
        c = Compound('Phosphate')
        self.assertIsNone(c.compartment)

    def test_compound_arguments(self):
        c = Compound('Menaquinone', arguments=[8])
        self.assertEqual(list(c.arguments), [8])

    def test_compound_arguments_none(self):
        c = Compound('Phosphate')
        self.assertEqual(list(c.arguments), [])

    def test_compound_translate_name(self):
        c = Compound('Pb')
        self.assertEqual(c.translate(lambda x: x.lower()), Compound('pb'))
        self.assertIsNot(c.translate(lambda x: x.lower()), c)

    def test_compound_in_compartment_when_unassigned(self):
        c = Compound('H+')
        self.assertEqual(c.in_compartment('e'), Compound('H+', 'e'))
        self.assertIsNot(c.in_compartment('e'), c)

    def test_compound_in_compartment_when_assigned(self):
        c = Compound('H+', 'c')
        self.assertEqual(c.in_compartment('p'), Compound('H+', 'p'))
        self.assertIsNot(c.in_compartment('p'), c)

    def test_compound_equals_other_with_same_name(self):
        c = Compound('Phosphate')
        self.assertEqual(c, Compound('Phosphate'))

    def test_compound_not_equals_other_with_name(self):
        c = Compound('H+')
        self.assertNotEqual(c, Compound('Phosphate'))

    def test_compound_equals_other_with_compartment(self):
        c = Compound('Phosphate', 'e')
        self.assertEqual(c, Compound('Phosphate', 'e'))

    def test_compound_not_equals_other_with_compartment(self):
        c = Compound('Phosphate', 'e')
        self.assertNotEqual(c, Compound('Phosphate', None))

    def test_compound_equals_other_with_arguments(self):
        c = Compound('Polyphosphate', arguments=[5])
        self.assertEqual(c, Compound('Polyphosphate', arguments=[5]))

    def test_compound_not_equals_other_with_arguments(self):
        c = Compound('Polyphosphate', arguments=[4])
        self.assertNotEqual(c, Compound('Polyphosphate', arguments=[5]))

    def test_compounds_sorted(self):
        l = sorted([Compound('A', arguments=[10]), Compound('A'), Compound('B'),
                    Compound('A', arguments=[4]), Compound('A', 'e')])
        self.assertEqual(l, [Compound('A'), Compound('A', arguments=[4]),
                              Compound('A', arguments=[10]), Compound('A', 'e'), Compound('B')])

    def test_compound_str_basic(self):
        self.assertEqual(str(Compound('Phosphate')), 'Phosphate')

    def test_compound_str_with_compartment(self):
        self.assertEqual(str(Compound('Phosphate', 'e')), 'Phosphate[e]')

    def test_compound_str_with_argument(self):
        self.assertEqual(str(Compound('Polyphosphate', arguments=[3])), 'Polyphosphate(3)')

    def test_compound_str_with_compartment_and_argument(self):
        self.assertEqual(str(Compound('Polyphosphate', 'p', [3])), 'Polyphosphate(3)[p]')


class TestReaction(unittest.TestCase):
    def test_reaction_init_empty_bidir(self):
        r = Reaction(Reaction.Bidir, [], [])

    def test_reaction_init_empty_left(self):
        r = Reaction(Reaction.Left, [], [])

    def test_reaction_init_empty_right(self):
        r = Reaction(Reaction.Right, [], [])

    def test_reaction_init_empty_invalid_direction(self):
        with self.assertRaises(ValueError):
            r = Reaction(None, [], [])

    def test_reaction_init_left_empty(self):
        r = Reaction(Reaction.Bidir, [], [(Compound('A'), 1)])

    def test_reaction_init_right_empty(self):
        r = Reaction(Reaction.Bidir, [(Compound('A'), 1)], [])

    def test_reaction_init_none_empty(self):
        r = Reaction(Reaction.Bidir, [(Compound('A'), 1)], [(Compound('B'), 1)])

    def test_reaction_direction_property(self):
        r = Reaction(Reaction.Bidir, [(Compound('A'), 1)], [(Compound('B'), 1)])
        self.assertEqual(r.direction, Reaction.Bidir)

    def test_reaction_left_property(self):
        r = Reaction(Reaction.Bidir, [(Compound('A'), 1)], [(Compound('B'), 1)])
        self.assertEqual(list(r.left), [(Compound('A'), 1)])

    def test_reaction_right_property(self):
        r = Reaction(Reaction.Bidir, [(Compound('A'), 1)], [(Compound('B'), 1)])
        self.assertEqual(list(r.right), [(Compound('B'), 1)])

    def test_reaction_compounds_property(self):
        r = Reaction(Reaction.Bidir, [(Compound('A'), 1)], [(Compound('B'), 1)])
        self.assertEqual(list(r.compounds),
                         [(Compound('A'), -1), (Compound('B'), 1)])

    def test_reaction_normalized_of_bidir(self):
        r = Reaction(Reaction.Bidir, [(Compound('Au'), 1)], [(Compound('Pb'), 1)]).normalized()
        self.assertEqual(r, Reaction(Reaction.Bidir, [(Compound('Au'), 1)], [(Compound('Pb'), 1)]))

    def test_reaction_normalized_of_right(self):
        r = Reaction(Reaction.Right, [(Compound('Au'), 1)], [(Compound('Pb'), 1)]).normalized()
        self.assertEqual(r, Reaction(Reaction.Right, [(Compound('Au'), 1)], [(Compound('Pb'), 1)]))

    def test_reaction_normalized_of_left(self):
        r = Reaction(Reaction.Left, [(Compound('Au'), 1)], [(Compound('Pb'), 1)]).normalized()
        self.assertEqual(r, Reaction(Reaction.Right, [(Compound('Pb'), 1)], [(Compound('Au'), 1)]))

    def test_reaction_translated_compounds(self):
        r = Reaction(Reaction.Right, [(Compound('Pb'), 1)], [(Compound('Au'), 1)])
        rt = r.translated_compounds(lambda name: name.lower())
        self.assertEqual(rt, Reaction(Reaction.Right, [(Compound('pb'), 1)], [(Compound('au'), 1)]))

    def test_reaction_format_simple(self):
        r = Reaction(Reaction.Right, [(Compound('Pb'), 1)], [(Compound('Au'), 1)])
        self.assertEqual(str(r), '|Pb| => |Au|')

    def test_reaction_format_with_arguments(self):
        pp1 = Compound('Polyphosphate', arguments=[4])
        pp2 = Compound('Polyphosphate', arguments=[5])
        r = Reaction(Reaction.Right, [(Compound('ATP'), 1), (pp1, 1)], [(Compound('ADP'), 1), (pp2, 1)])
        self.assertEqual(str(r), '|ATP| + |Polyphosphate(4)| => |ADP| + |Polyphosphate(5)|')

    def test_reaction_equals_other(self):
        r = Reaction(Reaction.Right, [(Compound('Pb'), 1)], [(Compound('Au'), 1)])
        self.assertEqual(r, Reaction(Reaction.Right, [(Compound('Pb'), 1)], [(Compound('Au'), 1)]))

    def test_reaction_not_equals_other_with_different_compounds(self):
        r = Reaction(Reaction.Right, [(Compound('Au'), 1)], [(Compound('Pb'), 1)])
        self.assertNotEqual(r, Reaction(Reaction.Right, [(Compound('Pb'), 1)], [(Compound('Au'), 1)]))

    def test_reaction_not_equals_other_with_different_direction(self):
        r = Reaction(Reaction.Right, [(Compound('Au'), 1)], [(Compound('Pb'), 1)])
        self.assertNotEqual(r, Reaction(Reaction.Left, [(Compound('Pb'), 1)], [(Compound('Au'), 1)]))


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2015  Jon Lund Steffensen <jon_steffensen@uri.edu>

import unittest

from psamm.database import DictDatabase, ChainedDatabase
from psamm.reaction import Compound, Reaction
from psamm.datasource.modelseed import parse_reaction


class TestMetabolicDatabase(unittest.TestCase):
    def setUp(self):
        self.database = DictDatabase()
        self.database.set_reaction('rxn_1', parse_reaction('=> (2) |A|'))
        self.database.set_reaction('rxn_2', parse_reaction('|A| <=> |B|'))
        self.database.set_reaction('rxn_3', parse_reaction('|A| => |D[e]|'))
        self.database.set_reaction('rxn_4', parse_reaction('|A| => |C|'))
        self.database.set_reaction('rxn_5', parse_reaction('|C| => |D[e]|'))
        self.database.set_reaction('rxn_6', parse_reaction('|D[e]| =>'))

    def test_reactions(self):
        self.assertEqual(
            set(self.database.reactions),
            {'rxn_1', 'rxn_2', 'rxn_3', 'rxn_4', 'rxn_5', 'rxn_6'})

    def test_compounds(self):
        self.assertEqual(
            set(self.database.compounds),
            {Compound('A'), Compound('B'), Compound('C'), Compound('D', 'e')})

    def test_compartments(self):
        self.assertEqual(set(self.database.compartments), {None, 'e'})

    def test_has_reaction_existing(self):
        self.assertTrue(self.database.has_reaction('rxn_3'))

    def test_has_reaction_not_existing(self):
        self.assertFalse(self.database.has_reaction('rxn_7'))

    def test_is_reversible_is_true(self):
        self.assertTrue(self.database.is_reversible('rxn_2'))

    def test_is_reversible_is_false(self):
        self.assertFalse(self.database.is_reversible('rxn_5'))

    def test_get_reaction_values(self):
        self.assertEqual(set(self.database.get_reaction_values('rxn_2')),
                            { (Compound('A'), -1), (Compound('B'), 1) })

    def test_get_compound_reactions(self):
        self.assertEqual(set(self.database.get_compound_reactions(Compound('A'))),
                            { 'rxn_1', 'rxn_2', 'rxn_3', 'rxn_4' })

    def test_reversible(self):
        self.assertEqual(set(self.database.reversible), { 'rxn_2' })

    def test_get_reaction(self):
        reaction = parse_reaction('|A| => |D[e]|')
        self.assertEqual(self.database.get_reaction('rxn_3'), reaction)

    def test_set_reaction_with_zero_coefficient(self):
        reaction = Reaction(Reaction.Bidir, [(Compound('A'), 1), (Compound('B'), 0)],
                                            [(Compound('C'), 1)])
        self.database.set_reaction('rxn_new', reaction)
        self.assertNotIn('rxn_new', self.database.get_compound_reactions(Compound('B')))

    def test_matrix_get_item(self):
        self.assertEqual(self.database.matrix[Compound('A'), 'rxn_1'], 2)
        self.assertEqual(self.database.matrix[Compound('A'), 'rxn_2'], -1)
        self.assertEqual(self.database.matrix[Compound('B'), 'rxn_2'], 1)
        self.assertEqual(self.database.matrix[Compound('A'), 'rxn_4'], -1)
        self.assertEqual(self.database.matrix[Compound('C'), 'rxn_4'], 1)
        self.assertEqual(self.database.matrix[Compound('C'), 'rxn_5'], -1)
        self.assertEqual(self.database.matrix[Compound('D', 'e'), 'rxn_5'], 1)

    def test_matrix_get_item_invalid_key(self):
        with self.assertRaises(KeyError):
            a = self.database.matrix[Compound('A'), 'rxn_5']
        with self.assertRaises(KeyError):
            b = self.database.matrix['rxn_1']

    def test_matrix_set_item_is_invalid(self):
        with self.assertRaises(TypeError):
            self.database.matrix[Compound('A'), 'rxn_1'] = 4

    def test_matrix_iter(self):
        matrix_keys = { (Compound('A'), 'rxn_1'),
                        (Compound('A'), 'rxn_2'),
                        (Compound('B'), 'rxn_2'),
                        (Compound('A'), 'rxn_3'),
                        (Compound('D', 'e'), 'rxn_3'),
                        (Compound('A'), 'rxn_4'),
                        (Compound('C'), 'rxn_4'),
                        (Compound('C'), 'rxn_5'),
                        (Compound('D', 'e'), 'rxn_5'),
                        (Compound('D', 'e'), 'rxn_6') }
        self.assertEqual(set(iter(self.database.matrix)), matrix_keys)

    def test_matrix_len(self):
        self.assertEqual(len(self.database.matrix), 10)

class TestChainedDatabase(unittest.TestCase):
    def setUp(self):
        database1 = DictDatabase()
        database1.set_reaction('rxn_1', parse_reaction('|A| => |B|'))
        database1.set_reaction('rxn_2', parse_reaction('|B| => |C| + |D|'))
        database1.set_reaction('rxn_3', parse_reaction('|D| <=> |E|'))
        database1.set_reaction('rxn_4', parse_reaction('|F| => |G|'))

        database2 = DictDatabase()
        database2.set_reaction('rxn_2', parse_reaction('|B| => |C|'))
        database2.set_reaction('rxn_3', parse_reaction('|C| => |D|'))
        database2.set_reaction('rxn_4', parse_reaction('|F| <=> |G|'))
        database2.set_reaction('rxn_5', parse_reaction('|G| + |I| <=> |H|'))

        self.database = ChainedDatabase(database2, database1)

    def test_has_reaction_in_lower_database(self):
        self.assertTrue(self.database.has_reaction('rxn_1'))

    def test_has_reaction_in_upper_new(self):
        self.assertTrue(self.database.has_reaction('rxn_5'))

    def test_has_reaction_in_upper_shadowing(self):
        self.assertTrue(self.database.has_reaction('rxn_3'))

    def test_is_reversible_in_lower_database(self):
        self.assertFalse(self.database.is_reversible('rxn_1'))

    def test_is_reversible_in_upper_new(self):
        self.assertTrue(self.database.is_reversible('rxn_5'))

    def test_is_reversible_in_upper_shadowing(self):
        self.assertFalse(self.database.is_reversible('rxn_3'))

    def test_get_compound_reactions_in_upper(self):
        reactions = set(self.database.get_compound_reactions(Compound('H')))
        self.assertEqual(reactions, { 'rxn_5' })

    def test_get_compound_reactions_in_lower(self):
        reactions = set(self.database.get_compound_reactions(Compound('A')))
        self.assertEqual(reactions, { 'rxn_1' })

    def test_get_compound_reactions_in_both(self):
        reactions = set(self.database.get_compound_reactions(Compound('B')))
        self.assertEqual(reactions, { 'rxn_1', 'rxn_2' })

    def test_get_compound_reactions_in_both_and_shadowed(self):
        reactions = set(self.database.get_compound_reactions(Compound('D')))
        self.assertEqual(reactions, { 'rxn_3' })


if __name__ == '__main__':
    unittest.main()



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2015  Jon Lund Steffensen <jon_steffensen@uri.edu>

import time
import logging

from ..command import Command, SolverCommandMixin, CommandError
from .. import fluxanalysis

from six.moves import range

logger = logging.getLogger(__name__)


class RobustnessCommand(SolverCommandMixin, Command):
    """Run robustness analysis on the model.

    Given a reaction to maximize and a reaction to vary,
    the robustness analysis will run FBA while fixing the
    reaction to vary at each iteration. The reaction will
    be fixed at the specified number of steps between the
    minimum and maximum flux value specified in the model.
    """

    @classmethod
    def init_parser(cls, parser):
        parser.add_argument(
            '--steps', metavar='N', type=int, default=10,
            help='Number of flux value steps for varying reaction')
        parser.add_argument(
            '--minimum', metavar='V', type=float,
            help='Minumum flux value of varying reacton')
        parser.add_argument(
            '--maximum', metavar='V', type=float,
            help='Maximum flux value of varying reacton')
        parser.add_argument(
            '--loop-removal', help='Select type of loop removal constraints',
            choices=['none', 'tfba', 'l1min'], default='none')
        parser.add_argument('--objective', help='Reaction to maximize')
        parser.add_argument(
            '--all-reaction-fluxes',
            help='Print reaction flux for all model reactions',
            action='store_true')
        parser.add_argument('varying', help='Reaction to vary')
        super(RobustnessCommand, cls).init_parser(parser)

    def run(self):
        """Run robustness command."""

        # Load compound information
        compound_name = {}
        for compound in self._model.parse_compounds():
            if 'name' in compound.properties:
                compound_name[compound.id] = compound.properties['name']
            elif compound.id not in compound_name:
                compound_name[compound.id] = compound.id

        if self._args.objective is not None:
            reaction = self._args.objective
        else:
            reaction = self._model.get_biomass_reaction()
            if reaction is None:
                raise CommandError('The biomass reaction was not specified')

        if not self._mm.has_reaction(reaction):
            raise CommandError('Specified reaction is not in model: {}'.format(
                reaction))

        varying_reaction = self._args.varying
        if not self._mm.has_reaction(varying_reaction):
            raise CommandError('Specified reaction is not in model: {}'.format(
                varying_reaction))

        steps = self._args.steps
        if steps <= 0:
            raise CommandError('Invalid number of steps: {}\n'.format(steps))

        loop_removal = self._args.loop_removal
        if loop_removal == 'tfba':
            solver = self._get_solver(integer=True)
        else:
            solver = self._get_solver()

        p = fluxanalysis.FluxBalanceProblem(self._mm, solver)

        if loop_removal == 'none':
            logger.info('Loop removal disabled; spurious loops are allowed')
            run_fba = p.maximize
        elif loop_removal == 'l1min':
            logger.info('Loop removal using L1 minimization')
            run_fba = p.max_min_l1
        elif loop_removal == 'tfba':
            logger.info('Loop removal using thermodynamic constraints')
            p.add_thermodynamic()
            run_fba = p.maximize
        else:
            raise CommandError('Invalid loop constraint mode: {}'.format(
                loop_removal))

        # Determine minimum and maximum flux for varying reaction
        if self._args.maximum is None:
            p.maximize(varying_reaction)
            flux_max = p.get_flux(varying_reaction)
        else:
            flux_max = self._args.maximum

        if self._args.minimum is None:
            p.maximize({varying_reaction: -1})
            flux_min = p.get_flux(varying_reaction)
        else:
            flux_min = self._args.minimum

        if flux_min > flux_max:
            raise CommandError('Invalid flux range: {}, {}\n'.format(
                flux_min, flux_max))

        logger.info('Varying {} in {} steps between {} and {}'.format(
            varying_reaction, steps, flux_min, flux_max))

        start_time = time.time()

        # Run FBA on model at different fixed flux values
        for i in range(steps):
            fixed_flux = flux_min + i*(flux_max - flux_min)/float(steps-1)
            flux_var = p.get_flux_var(varying_reaction)
            c, = p.prob.add_linear_constraints(flux_var == fixed_flux)

            try:
                run_fba(reaction)

                if not self._args.all_reaction_fluxes:
                    print('{}\t{}'.format(fixed_flux, p.get_flux(reaction)))
                else:
                    for other_reaction in self._mm.reactions:
                        print('{}\t{}\t{}'.format(
                            other_reaction, fixed_flux,
                            p.get_flux(other_reaction)))
            except fluxanalysis.FluxBalanceError:
                logger.warning('No solution found for {} at {}'.format(
                    varying_reaction, fixed_flux))
            finally:
                c.delete()

        logger.info('Solving took {:.2f} seconds'.format(
            time.time() - start_time))



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2015  Jon Lund Steffensen <jon_steffensen@uri.edu>

"""Representation of metabolic network databases."""

import abc
from collections import defaultdict, Mapping

from six import iteritems, add_metaclass

from .reaction import Reaction


class StoichiometricMatrixView(Mapping):
    """Provides a sparse matrix view on the stoichiometry of a database.

    This object is used internally in the database to expose a sparse matrix
    view of the model stoichiometry. This class should not be instantied,
    instead use the :attr:`MetabolicDatabase.matrix` property. Any compound,
    reaction-pair can be looked up to obtain the corresponding stoichiometric
    value. If the value is not defined (implicitly zero) a
    :exc:`KeyError <exceptions.KeyError>` will be raised.

    In addition, instances also support the NumPy `__array__` protocol which
    means that a :class:`numpy.array` can by created directly from the matrix.

    >>> model = MetabolicModel()
    >>> matrix = numpy.array(model.matrix)
    """

    def __init__(self, database):
        super(StoichiometricMatrixView, self).__init__()
        self._database = database

    def __getitem__(self, key):
        if len(key) != 2:
            raise KeyError(key)
        compound, reaction = key
        if not self._database.has_reaction(reaction):
            raise KeyError(key)
        reaction_values = dict(self._database.get_reaction_values(reaction))
        if compound not in reaction_values:
            raise KeyError(key)
        return reaction_values[compound]

    def __iter__(self):
        for reaction in self._database.reactions:
            for compound, _ in self._database.get_reaction_values(reaction):
                yield compound, reaction

    def __len__(self):
        return sum(sum(1 for _ in self._database.get_reaction_values(reaction))
                   for reaction in self._database.reactions)

    def __array__(self):
        """Return Numpy ndarray instance of matrix.

        The matrix is indexed by sorted compound, reaction-keys
        """
        import numpy  # NumPy is only required for this method

        compound_list = sorted(self._database.compounds)
        reaction_list = sorted(self._database.reactions)
        matrix = numpy.zeros((len(compound_list), len(reaction_list)))

        compound_map = dict(
            (compound, i) for i, compound in enumerate(compound_list))
        reaction_map = dict(
            (reaction_id, i) for i, reaction_id in enumerate(reaction_list))

        for reaction_id in self._database.reactions:
            for compound, value in self._database.get_reaction_values(
                    reaction_id):
                c_index = compound_map[compound]
                r_index = reaction_map[reaction_id]
                matrix[c_index, r_index] = value

        return matrix


@add_metaclass(abc.ABCMeta)
class MetabolicDatabase(object):
    """Database of metabolic reactions."""

    @abc.abstractproperty
    def reactions(self):
        """Iterator of reactions IDs in the database."""

    @abc.abstractproperty
    def compounds(self):
        """Itertor of :class:`Compounds <psamm.reaction.Compound>` in the
        database."""

    @abc.abstractproperty
    def compartments(self):
        """Iterator of compartment IDs in the database."""

    @abc.abstractmethod
    def has_reaction(self, reaction_id):
        """Whether the given reaction exists in the database."""

    @abc.abstractmethod
    def is_reversible(self, reaction_id):
        """Whether the given reaction is reversible."""

    @abc.abstractmethod
    def get_reaction_values(self, reaction_id):
        """Return an iterator of reaction compounds and stoichiometric values.

        The returned iterator contains
        (:class:`Compound <psamm.reaction.Compound>`, value)-tuples. The value
        is negative for left-hand side compounds and positive for right-hand
        side.
        """

    @abc.abstractmethod
    def get_compound_reactions(self, compound_id):
        """Return an iterator of reactions containing the compound.

        Reactions are returned as IDs.
        """

    @property
    def reversible(self):
        """The set of reversible reactions."""
        return set(reaction_id for reaction_id in self.reactions
                   if self.is_reversible(reaction_id))

    @property
    def matrix(self):
        """Mapping from compound, reaction to stoichiometric value.

        This is an instance of :class:`StoichiometricMatrixView`."""
        return StoichiometricMatrixView(self)

    def get_reaction(self, reaction_id):
        """Return reaction as a :class:`Reaction <psamm.reaction.Reaction>`."""

        direction = Reaction.Right
        if self.is_reversible(reaction_id):
            direction = Reaction.Bidir

        left = ((compound, -value) for compound, value in
                self.get_reaction_values(reaction_id) if value < 0)
        right = ((compound, value) for compound, value in
                 self.get_reaction_values(reaction_id) if value > 0)

        return Reaction(direction, left, right)


class DictDatabase(MetabolicDatabase):
    """Metabolic database backed by in-memory dictionaries

    This is a subclass of :class:`MetabolicDatabase`."""

    def __init__(self):
        super(DictDatabase, self).__init__()
        self._reactions = defaultdict(dict)
        self._compound_reactions = defaultdict(set)
        self._reversible = set()

    @property
    def reactions(self):
        return iter(self._reactions)

    @property
    def compounds(self):
        return iter(self._compound_reactions)

    @property
    def compartments(self):
        compartment_set = set()
        for compound in self.compounds:
            if compound.compartment not in compartment_set:
                compartment_set.add(compound.compartment)
                yield compound.compartment

    def has_reaction(self, reaction_id):
        return reaction_id in self._reactions

    def is_reversible(self, reaction_id):
        return reaction_id in self._reversible

    def get_reaction_values(self, reaction_id):
        if reaction_id not in self._reactions:
            raise ValueError('Unknown reaction: {}'.format(repr(reaction_id)))
        return iteritems(self._reactions[reaction_id])

    def get_compound_reactions(self, compound_id):
        return iter(self._compound_reactions[compound_id])

    def set_reaction(self, reaction_id, reaction):
        """Set the reaction ID to a reaction given by a
        :class:`Reaction <psamm.reaction.Reaction>`

        If an existing reaction exists with the given reaction ID it will be
        overwritten.
        """

        # Overwrite previous reaction if the same id is used
        if reaction_id in self._reactions:
            # Clean up compound to reaction mapping
            for compound in self._reactions[reaction_id]:
                self._compound_reactions[compound].remove(reaction_id)

            self._reversible.discard(reaction_id)
            del self._reactions[reaction_id]

        # Add values to global (sparse) stoichiometric matrix
        # Compounds that occur on both sides will get a stoichiometric
        # value based on the sum of the signed values on each side.
        for compound, _ in reaction.compounds:
            if compound not in self._reactions[reaction_id]:
                self._reactions[reaction_id][compound] = 0
                self._compound_reactions[compound].add(reaction_id)
        for compound, value in reaction.left:
            self._reactions[reaction_id][compound] -= value
        for compound, value in reaction.right:
            self._reactions[reaction_id][compound] += value

        # Remove reaction from compound reactions if the resulting
        # stoichiometric value turned out to be zero.
        zero_compounds = set()
        for compound, value in iteritems(self._reactions[reaction_id]):
            if value == 0:
                zero_compounds.add(compound)

        for compound in zero_compounds:
            del self._reactions[reaction_id][compound]
            self._compound_reactions[compound].remove(reaction_id)

        if reaction.direction != '=>':
            self._reversible.add(reaction_id)


class ChainedDatabase(MetabolicDatabase):
    """Links a number of databases so they can be treated a single database

    This is a subclass of :class:`MetabolicDatabase`."""

    def __init__(self, *databases):
        self._databases = list(databases)
        if len(self._databases) == 0:
            self._databases.append(DictDatabase())

    def _is_shadowed(self, reaction_id, database):
        """Whether reaction in database is shadowed by another database"""
        for other_database in self._databases:
            if other_database == database:
                break
            if other_database.has_reaction(reaction_id):
                return True
        return False

    @property
    def reactions(self):
        # Make sure that we only yield each reaction once
        reaction_set = set()
        for database in self._databases:
            for reaction in database.reactions:
                if reaction not in reaction_set:
                    reaction_set.add(reaction)
                    yield reaction

    @property
    def compounds(self):
        # Make sure that we only yield each compound once
        compound_set = set()
        for database in self._databases:
            for compound in database.compounds:
                if compound not in compound_set:
                    compound_set.add(compound)
                    yield compound

    @property
    def compartments(self):
        # Make sure that we yield each compartment once
        compartment_set = set()
        for database in self._databases:
            for compartment in database.compartments:
                if compartment not in compartment_set:
                    compartment_set.add(compartment)
                    yield compartment

    def has_reaction(self, reaction_id):
        return any(database.has_reaction(reaction_id)
                   for database in self._databases)

    def is_reversible(self, reaction_id):
        for database in self._databases:
            if database.has_reaction(reaction_id):
                return database.is_reversible(reaction_id)
        raise ValueError('Unknown reaction: {}'.format(reaction_id))

    def get_reaction_values(self, reaction_id):
        for database in self._databases:
            if database.has_reaction(reaction_id):
                return database.get_reaction_values(reaction_id)
        raise ValueError('Unknown reaction: {}'.format(reaction_id))

    def get_compound_reactions(self, compound):
        # Make sure that we only yield each reaction once
        reaction_set = set()
        for database in self._databases:
            for reaction in database.get_compound_reactions(compound):
                if (reaction not in reaction_set and
                        not self._is_shadowed(reaction, database)):
                    reaction_set.add(reaction)
                    yield reaction

    def set_reaction(self, reaction_id, reaction):
        if hasattr(self._databases[0], 'set_reaction'):
            self._databases[0].set_reaction(reaction_id, reaction)
        else:
            raise ValueError('First database is immutable')



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2015  Jon Lund Steffensen <jon_steffensen@uri.edu>

"""Linear programming solver using QSopt_ex."""

from __future__ import absolute_import

from itertools import repeat, count
import numbers

from six import iteritems
from six.moves import zip
import qsoptex

from .lp import Solver as BaseSolver
from .lp import Constraint as BaseConstraint
from .lp import Problem as BaseProblem
from .lp import Result as BaseResult
from .lp import (Expression, Relation, ObjectiveSense, VariableType,
                 InvalidResultError)


class Solver(BaseSolver):
    """Represents an LP solver using QSopt_ex"""

    def create_problem(self, **kwargs):
        """Create a new LP-problem using the solver"""
        return Problem(**kwargs)


class Problem(BaseProblem):
    """Represents an LP-problem of a qsoptex.Solver"""

    CONSTR_SENSE_MAP = {
        Relation.Equals: qsoptex.ConstraintSense.EQUAL,
        Relation.Greater: qsoptex.ConstraintSense.GREATER,
        Relation.Less: qsoptex.ConstraintSense.LESS
    }

    def __init__(self, **kwargs):
        self._p = qsoptex.ExactProblem()
        self._p.set_param(qsoptex.Parameter.SIMPLEX_DISPLAY, 1)

        self._variables = {}
        self._var_names = ('x'+str(i) for i in count(1))
        self._constr_names = ('c'+str(i) for i in count(1))

        self._result = None

    @property
    def qsoptex(self):
        """The underlying qsoptex.ExactProblem object"""
        return self._p

    def define(self, *names, **kwargs):
        """Define variable in the problem

        Variables must be defined before they can be accessed by var() or
        set(). This function takes keyword arguments lower and upper to define
        the bounds of the variable (default: -inf to inf). The keyword argument
        types can be used to select the type of the variable (Only Continuous
        is suported).
        """
        names = tuple(names)
        lower = kwargs.get('lower', None)
        upper = kwargs.get('upper', None)
        vartype = kwargs.get('types', None)

        # Repeat values if a scalar is given
        if lower is None or isinstance(lower, numbers.Number):
            lower = repeat(lower, len(names))
        if upper is None or isinstance(upper, numbers.Number):
            upper = repeat(upper, len(names))
        if vartype is None or vartype in (
                VariableType.Continuous, VariableType.Binary,
                VariableType.Integer):
            vartype = repeat(vartype, len(names))

        lp_names = tuple(next(self._var_names) for name in names)

        # Assign default values
        vartype = (VariableType.Continuous if value is None else value
                   for value in vartype)

        self._variables.update(zip(names, lp_names))
        for name, lower, upper, t in zip(lp_names, lower, upper, vartype):
            if t != VariableType.Continuous:
                raise ValueError(
                    'Solver does not support non-continuous types')
            self._p.add_variable(0, lower, upper, name)

    def has_variable(self, name):
        """Check whether variable is defined in the model."""
        return name in self._variables

    def _add_constraints(self, relation):
        """Add the given relation as one or more constraints

        Return a list of the names of the constraints added.
        """
        expression = relation.expression
        names = []
        for value_set in expression.value_sets():
            values = ((self._variables[variable], value)
                      for variable, value in value_set)
            constr_name = next(self._constr_names)
            sense = self.CONSTR_SENSE_MAP[relation.sense]
            self._p.add_linear_constraint(
                sense=sense, values=values, rhs=-expression.offset,
                name=constr_name)
            names.append(constr_name)

        return names

    def add_linear_constraints(self, *relations):
        """Add constraints to the problem

        Each constraint is represented by a Relation, and the
        expression in that relation can be a set expression.
        """
        constraints = []

        for relation in relations:
            if self._check_relation(relation):
                constraints.append(Constraint(self, None))
            else:
                for name in self._add_constraints(relation):
                    constraints.append(Constraint(self, name))

        return constraints

    def set_linear_objective(self, expression):
        """Set linear objective of problem"""

        if isinstance(expression, numbers.Number):
            # Allow expressions with no variables as objective,
            # represented as a number
            expression = Expression()

        self._p.set_linear_objective(
            (lp_name, expression.value(var))
            for var, lp_name in iteritems(self._variables))

    def set_objective_sense(self, sense):
        """Set type of problem (maximize or minimize)"""
        if sense == ObjectiveSense.Minimize:
            self._p.set_objective_sense(qsoptex.ObjectiveSense.MINIMIZE)
        elif sense == ObjectiveSense.Maximize:
            self._p.set_objective_sense(qsoptex.ObjectiveSense.MAXIMIZE)
        else:
            raise ValueError('Invalid objective sense')

    def solve(self, sense=None):
        """Solve problem"""
        if sense is not None:
            self.set_objective_sense(sense)
        self._p.solve()

        self._result = Result(self)
        return self._result

    @property
    def result(self):
        return self._result


class Constraint(BaseConstraint):
    """Represents a constraint in a qsoptex.Problem"""

    def __init__(self, prob, name):
        self._prob = prob
        self._name = name

    def delete(self):
        if self._name is not None:
            self._prob._p.delete_linear_constraint(self._name)


class Result(BaseResult):
    """Represents the solution to a qsoptex.Problem

    This object will be returned from the Problem.solve() method or by
    accessing the Problem.result property after solving a problem. This
    class should not be instantiated manually.

    Result will evaluate to a boolean according to the success of the
    solution, so checking the truth value of the result will immediately
    indicate whether solving was successful.
    """

    def __init__(self, prob):
        self._problem = prob

    def _check_valid(self):
        if self._problem.result != self:
            raise InvalidResultError()

    @property
    def success(self):
        """Return boolean indicating whether a solution was found"""
        self._check_valid()
        return self._problem._p.get_status() == qsoptex.SolutionStatus.OPTIMAL

    @property
    def unbounded(self):
        """Whether the solution is unbounded"""
        self._check_valid()
        return (self._problem._p.get_status() ==
                qsoptex.SolutionStatus.UNBOUNDED)

    @property
    def status(self):
        """Return string indicating the error encountered on failure"""
        self._check_valid()
        return 'Status: {}'.format(self._problem._p.get_status())

    def get_value(self, expression):
        """Return value of expression"""

        self._check_valid()
        if isinstance(expression, Expression):
            return sum(self._problem._p.get_value(
                self._problem._variables[var])*value
                for var, value in expression.values())
        elif expression not in self._problem._variables:
            raise ValueError('Unknown expression: {}'.format(expression))
        return self._problem._p.get_value(self._problem._variables[expression])



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2015  Jon Lund Steffensen <jon_steffensen@uri.edu>

"""Generic interface to LP solver instantiation."""

from __future__ import absolute_import, print_function

import argparse
import os
import logging

from .lp import Solver as BaseSolver

from six import iteritems

logger = logging.getLogger(__name__)
_solvers = []
_solver_import_errors = {}


# Try to load Cplex solver
try:
    from . import cplex
    _solvers.append({
        'class': cplex.Solver,
        'name': 'cplex',
        'integer': True,
        'rational': False,
        'priority': 10
    })
except ImportError as e:
    _solver_import_errors['cplex'] = str(e)

# Try to load QSopt_ex solver
try:
    from . import qsoptex
    _solvers.append({
        'class': qsoptex.Solver,
        'name': 'qsoptex',
        'integer': False,
        'rational': True,
        'priority': 5
    })
except ImportError as e:
    _solver_import_errors['qsoptex'] = str(e)

# Try to load Gurobi solver
try:
    from . import gurobi
    _solvers.append({
        'class': gurobi.Solver,
        'name': 'gurobi',
        'integer': True,
        'rational': False,
        'priority': 9
    })
except ImportError as e:
    _solver_import_errors['gurobi'] = str(e)


class RequirementsError(Exception):
    """Error resolving solver requirements"""


def filter_solvers(solvers, requirements):
    """Yield solvers that fullfil the requirements."""
    for solver in solvers:
        for req, value in iteritems(requirements):
            if (req in ('integer', 'rational', 'name') and
                    (req not in solver or solver[req] != value)):
                break
        else:
            yield solver


class Solver(BaseSolver):
    """Generic solver interface based on requirements

    Use the any of the following keyword arguments to restrict which
    underlying solver is used:

    - `integer`: Use a solver that support integer variables (MILP)
    - `rational`: Use a solver that returns rational results
    - `name`: Select a specific solver based on the name
    """

    def __init__(self, **kwargs):
        if len(_solvers) == 0:
            raise RequirementsError('No solvers available')

        self._requirements = {key: value for key, value in iteritems(kwargs)
                              if value is not None}
        solvers = list(filter_solvers(_solvers, self._requirements))

        # Obtain solver priority from environment variable, if specified.
        priority = {}
        if 'PSAMM_SOLVER' in os.environ:
            solver_names = os.environ['PSAMM_SOLVER'].split(',')
            for i, solver_name in enumerate(solver_names):
                priority[solver_name] = len(solver_names) - i
            solvers = [s for s in solvers if s['name'] in priority]
        else:
            # Use built-in priorities
            for solver in solvers:
                priority[solver['name']] = solver['priority']

        if len(solvers) == 0:
            raise RequirementsError(
                'Unable to find a solver matching the specified requirements:'
                ' {}'.format(self._requirements))

        solver = max(solvers, key=lambda s: priority.get(s['name'], 0))
        logger.debug('Using solver {}'.format(solver['name']))

        self._properties = solver
        self._solver = solver['class']()

    @property
    def properties(self):
        return self._properties

    def create_problem(self):
        """Create a :class:`Problem <psamm.lpsolver.lp.Problem>` instance"""
        return self._solver.create_problem(**self._requirements)


def parse_solver_setting(s):
    """Parse a string containing a solver setting"""

    try:
        key, value = s.split('=', 1)
    except ValueError:
        key, value = s, 'yes'

    if key in ('rational', 'integer'):
        value = value.lower() in ('1', 'yes', 'true', 'on')
    elif key in ('threads',):
        value = int(value)
    elif key in ('feasibility_tolerance',):
        value = float(value)

    return key, value


def list_solvers(args=None):
    """Entry point for listing available solvers."""
    parser = argparse.ArgumentParser(
        description='''List LP solver available in PSAMM. This will produce a
                       list of all of the available LP solvers in prioritized
                       order. Addtional requirements can be imposed with the
                       arguments (e.g. integer=yes to select only solvers that
                       support MILP problems). The list will also be influenced
                       by the PSAMM_SOLVER environment variable which can be
                       used to only allow specific solvers (e.g.
                       PSAMM_SOLVER=cplex).''')
    parser.add_argument(
        'requirement', nargs='*', type=str,
        help='Additional requirements on the selected solvers')
    parsed_args = parser.parse_args(args)

    requirements = {}
    for arg in parsed_args.requirement:
        try:
            key, value = parse_solver_setting(arg)
        except ValueError as e:
            parser.error(str(e))
        else:
            requirements[key] = value

    solvers = list(filter_solvers(_solvers, requirements))
    solver_names = set(solver['name'] for solver in solvers)

    # Obtain solver priority from environment variable, if specified.
    priority = {}
    if 'PSAMM_SOLVER' in os.environ:
        names = os.environ['PSAMM_SOLVER'].split(',')
        for i, solver_name in enumerate(names):
            priority[solver_name] = len(names) - i
        solvers = [s for s in solvers if s['name'] in priority]
        solver_names = set(priority)
    else:
        # Use built-in priorities
        for solver in solvers:
            priority[solver['name']] = solver['priority']

    solvers = sorted(solvers, key=lambda s: priority.get(s['name'], 0),
                     reverse=True)

    status = 0

    if len(solvers) > 0:
        print('Prioritized solvers:')
        for solver in solvers:
            print('Name: {}'.format(solver['name']))
            print('Priority: {}'.format(solver['priority']))
            print('MILP (integer) problem support: {}'.format(
                solver['integer']))
            print('Rational solution: {}'.format(solver['rational']))
            print('Class: {}'.format(solver['class']))
            print()
    else:
        status = 1
        print('No solvers fullfil the requirements!')
        print()

    filtered_solvers_count = len(_solvers) - len(solvers)
    if filtered_solvers_count > 0 or len(_solver_import_errors) > 0:
        print('Unavailable solvers:')
        for solver in _solvers:
            if solver['name'] not in solver_names:
                print('{}: Does not fullfil the specified requirements'.format(
                    solver['name']))

        for solver, error in iteritems(_solver_import_errors):
            print('{}: Error loading solver: {}'.format(solver, error))

    if status != 0:
        parser.exit(status)



#!/usr/bin/env python
# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2015  Jon Lund Steffensen <jon_steffensen@uri.edu>

import os
import unittest

from psamm.lpsolver import generic


class TestListSolversCommand(unittest.TestCase):
    def test_list_lpsolvers(self):
        if os.environ.get('PSAMM_SOLVER') in ('nosolver', None):
            with self.assertRaises(SystemExit):
                generic.list_solvers([])
        else:
            generic.list_solvers([])



#!/usr/bin/env python
# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2015  Jon Lund Steffensen <jon_steffensen@uri.edu>

import unittest

from psamm.metabolicmodel import MetabolicModel
from psamm.database import DictDatabase
from psamm import massconsistency
from psamm.datasource.modelseed import parse_reaction
from psamm.reaction import Compound
from psamm.lpsolver import generic

from six import iteritems


class TestMassConsistency(unittest.TestCase):
    """Test mass consistency using a simple model"""

    def setUp(self):
        # TODO use mock model instead of actual model
        self.database = DictDatabase()
        self.database.set_reaction('rxn_1', parse_reaction('=> (2) |A|'))
        self.database.set_reaction('rxn_2', parse_reaction('|A| <=> |B|'))
        self.database.set_reaction('rxn_3', parse_reaction('|A| => |D|'))
        self.database.set_reaction('rxn_4', parse_reaction('|A| => |C|'))
        self.database.set_reaction('rxn_5', parse_reaction('|C| => |D|'))
        self.database.set_reaction('rxn_6', parse_reaction('|D| =>'))
        self.model = MetabolicModel.load_model(self.database, self.database.reactions)

        try:
            self.solver = generic.Solver()
        except generic.RequirementsError:
            self.skipTest('Unable to find an LP solver for tests')

    def test_mass_consistent_is_consistent(self):
        exchange = { 'rxn_1', 'rxn_6' }
        self.assertTrue(massconsistency.is_consistent(
            self.model, self.solver, exchange, set()))

    def test_mass_inconsistent_is_consistent(self):
        exchange = { 'rxn_1', 'rxn_6' }
        self.database.set_reaction('rxn_7', parse_reaction('|D| => (2) |C|'))
        self.model.add_reaction('rxn_7')
        self.assertFalse(massconsistency.is_consistent(
            self.model, self.solver, exchange, set()))

    def test_mass_consistent_reactions_returns_compounds(self):
        exchange = { 'rxn_1', 'rxn_6' }
        _, compounds = massconsistency.check_reaction_consistency(
            self.model, exchange=exchange, solver=self.solver)
        for c, value in compounds:
            self.assertIn(c, self.model.compounds)
            self.assertGreaterEqual(value, 1.0)

    def test_mass_consistent_reactions_returns_reactions(self):
        exchange = { 'rxn_1', 'rxn_6' }
        reactions, _ = massconsistency.check_reaction_consistency(
            self.model, exchange=exchange, solver=self.solver)
        for r, residual in reactions:
            self.assertIn(r, self.model.reactions)


class TestMassConsistencyZeroMass(unittest.TestCase):
    """Test mass consistency using a model with zero-mass compound"""

    def setUp(self):
        # TODO use mock model instead of actual model
        self.database = DictDatabase()
        self.database.set_reaction('rxn_1', parse_reaction(
            '|A| + |B| => |C|'))
        self.database.set_reaction('rxn_2', parse_reaction(
            '|C| + |Z| => |A| + |B|'))
        self.model = MetabolicModel.load_model(
            self.database, self.database.reactions)

        try:
            self.solver = generic.Solver()
        except generic.RequirementsError:
            self.skipTest('Unable to find an LP solver for tests')

    def test_is_consistent_with_zeromass(self):
        consistent = massconsistency.is_consistent(
            self.model, solver=self.solver, zeromass={Compound('Z')})
        self.assertTrue(consistent)

    def test_compound_consistency_with_zeromass(self):
        compounds = dict(massconsistency.check_compound_consistency(
            self.model, solver=self.solver, zeromass={Compound('Z')}))
        self.assertEquals(compounds[Compound('Z')], 0)
        for c, value in iteritems(compounds):
            if c.name != 'Z':
                self.assertGreaterEqual(value, 1)

    def test_reaction_consistency_with_zeromass(self):
        reactions, _ = massconsistency.check_reaction_consistency(
            self.model, solver=self.solver, zeromass={Compound('Z')})
        reactions = dict(reactions)

        for r, value in iteritems(reactions):
            self.assertAlmostEqual(value, 0)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2015  Jon Lund Steffensen <jon_steffensen@uri.edu>

import os
import shutil
import tempfile
import unittest

from psamm.datasource import native
from psamm.reaction import Reaction, Compound

from six import StringIO


class TestYAMLDataSource(unittest.TestCase):
    def test_parse_reaction_list(self):
        reactions = list(native.parse_reaction_list('./test.yaml', [
            {
                'id': 'rxn1',
                'equation': {
                    'reversible': True,
                    'left': [
                        { 'id': 'A', 'value': 1 },
                        { 'id': 'B', 'value': 2 } ],
                    'right': [
                        { 'id': 'C', 'value': 1 }
                    ]
                }
            }
        ]))

        self.assertEqual(len(reactions), 1)

        reaction = Reaction(Reaction.Bidir,
                            [(Compound('A'), 1), (Compound('B'), 2)],
                            [(Compound('C'), 1)])
        self.assertEqual(reactions[0].equation, reaction)

    def test_parse_reaction_list_missing_value(self):
        with self.assertRaises(native.ParseError):
            reactions = list(native.parse_reaction_list('./test.yaml', [
                {
                    'id': 'rxn1',
                    'equation': {
                        'left': [
                            { 'id': 'A' }
                        ]
                    }
                }
            ]))

    def test_parse_medium_table(self):
        table = '''
ac      e
glcD    e       -10
co2     e       -       50
'''

        medium = list(native.parse_medium_table_file(StringIO(table.strip())))
        self.assertEqual(len(medium), 3)
        self.assertEqual(medium[0], (Compound('ac', 'e'), None, None, None))
        self.assertEqual(medium[1], (Compound('glcD', 'e'), None, -10, None))
        self.assertEqual(medium[2], (Compound('co2', 'e'), None, None, 50))

    def test_parse_medium(self):
        medium = list(native.parse_medium({
            'compartment': 'e',
            'compounds': [
                {'id': 'ac'},
                {'id': 'glcD', 'lower': -10},
                {'id': 'co2', 'upper': 50},
                {'id': 'compound_x', 'compartment': 'c'},
                {'id': 'compound_y', 'reaction': 'EX_cpdy'}
            ]
        }))

        self.assertEqual(len(medium), 5)
        self.assertEqual(medium[0], (Compound('ac', 'e'), None, None, None))
        self.assertEqual(medium[1], (Compound('glcD', 'e'), None, -10, None))
        self.assertEqual(medium[2], (Compound('co2', 'e'), None, None, 50))
        self.assertEqual(
            medium[3], (Compound('compound_x', 'c'), None, None, None))
        self.assertEqual(
            medium[4], (Compound('compound_y', 'e'), 'EX_cpdy', None, None))


class TestYAMLFileSystemData(unittest.TestCase):
    """Test loading files from file system."""

    def setUp(self):
        self._model_dir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self._model_dir)

    def write_model_file(self, filename, contents):
        path = os.path.join(self._model_dir, filename)
        with open(path, 'w') as f:
            f.write(contents)
        return path

    def test_parse_compound_tsv_file(self):
        path = self.write_model_file('compounds.tsv', '\n'.join([
            'id\tname\tformula\tcharge\tkegg\tcas',
            'h2o\tWater\tH2O\t0\tC00001\t7732-18-5'
        ]))

        compounds = list(native.parse_compound_file(path, 'tsv'))
        self.assertEqual(len(compounds), 1)

        self.assertEqual(compounds[0].id, 'h2o')
        self.assertEqual(compounds[0].name, 'Water')
        self.assertEqual(compounds[0].properties['name'], 'Water')
        self.assertEqual(compounds[0].formula, 'H2O')
        self.assertEqual(compounds[0].charge, 0)
        self.assertEqual(compounds[0].kegg, 'C00001')
        self.assertEqual(compounds[0].cas, '7732-18-5')

    def test_parse_compound_yaml_file(self):
        path = self.write_model_file('compounds.yaml', '\n'.join([
            '- id: h2o',
            '  name: Water',
            '  formula: H2O',
            '  user_annotation: ABC'
        ]))

        compounds = list(native.parse_compound_file(path, 'yaml'))
        self.assertEqual(len(compounds), 1)

        self.assertEqual(compounds[0].id, 'h2o')
        self.assertEqual(compounds[0].name, 'Water')
        self.assertEqual(compounds[0].formula, 'H2O')
        self.assertEqual(compounds[0].properties['user_annotation'], 'ABC')

    def test_parse_reaction_table_file(self):
        path = self.write_model_file('reactions.tsv', '\n'.join([
            'id\tname\tequation',
            'rxn_1\tReaction 1\t|A| => (2) |B|',
            'rxn_2\tSecond reaction\t<=> |C|'
        ]))

        reactions = list(native.parse_reaction_file(path))
        self.assertEqual(len(reactions), 2)

        self.assertEqual(reactions[0].id, 'rxn_1')
        self.assertEqual(reactions[0].name, 'Reaction 1')
        self.assertEqual(reactions[0].properties['name'], 'Reaction 1')
        self.assertEqual(reactions[0].equation, Reaction(
            Reaction.Right, [(Compound('A'), 1)], [(Compound('B'), 2)]))
        self.assertEqual(reactions[0].ec, None)
        self.assertEqual(reactions[0].genes, None)
        self.assertEqual(reactions[0].filemark.filecontext.filepath, path)
        self.assertEqual(reactions[0].filemark.line, 2)

        self.assertEqual(reactions[1].id, 'rxn_2')
        self.assertEqual(reactions[1].name, 'Second reaction')
        self.assertEqual(reactions[1].equation, Reaction(
            Reaction.Bidir, [], [(Compound('C'), 1)]))
        self.assertEqual(reactions[1].filemark.filecontext.filepath, path)
        self.assertEqual(reactions[1].filemark.line, 3)

    def test_parse_reaction_yaml_file(self):
        path = self.write_model_file('reaction.yaml', '\n'.join([
            '- id: rxn_1',
            '  equation: "|A| => |B|"',
            '- id: rxn_2',
            '  equation:',
            '    reversible: no',
            '    left:',
            '      - id: B',
            '        value: 1',
            '    right:',
            '      - id: A',
            '        value: 1'
        ]))

        reactions = list(native.parse_reaction_file(path))
        self.assertEqual(len(reactions), 2)

        self.assertEqual(reactions[0].id, 'rxn_1')
        self.assertEqual(reactions[1].id, 'rxn_2')

    def test_parse_medium_table_file(self):
        path = self.write_model_file('medium.tsv', '\n'.join([
            '',
            '# comment',
            'cpd_A\tc',
            'cpd_B\te\t-1000',
            'cpd_C\te\t-\t20  # line comment',
            'cpd_D\te\t-100\t-10'
        ]))

        medium = list(native.parse_medium_file(path))
        self.assertEqual(medium, [
            (Compound('cpd_A', 'c'), None, None, None),
            (Compound('cpd_B', 'e'), None, -1000, None),
            (Compound('cpd_C', 'e'), None, None, 20),
            (Compound('cpd_D', 'e'), None, -100, -10)
        ])

    def test_parse_medium_yaml_file(self):
        path = self.write_model_file('medium.yaml', '\n'.join([
            'compartment: e',
            'compounds:',
            '  - id: cpd_A',
            '    reaction: EX_A',
            '    lower: -40',
            '  - id: cpd_B',
            '    upper: 100',
            '  - id: cpd_C',
            '    lower: -100.0',
            '    upper: 500.0',
            '  - id: cpd_D',
            '    compartment: c'
        ]))

        medium = list(native.parse_medium_file(path))
        self.assertEqual(medium, [
            (Compound('cpd_A', 'e'), 'EX_A', -40, None),
            (Compound('cpd_B', 'e'), None, None, 100),
            (Compound('cpd_C', 'e'), None, -100, 500),
            (Compound('cpd_D', 'c'), None, None, None)
        ])

    def test_parse_medium_yaml_list(self):
        self.write_model_file('medium.yaml', '\n'.join([
            'compartment: e',
            'compounds:',
            '  - id: cpd_A',
            '    lower: -42'
        ]))

        path = os.path.join(self._model_dir, 'fake.yaml')
        medium = list(native.parse_medium_list(path, [
            {'include': 'medium.yaml'},
            {
                'compartment': 'e',
                'compounds': [
                    {'id': 'cpd_B', 'upper': 767}
                ]
            }
        ]))

        self.assertEqual(medium, [
            (Compound('cpd_A', 'e'), None, -42, None),
            (Compound('cpd_B', 'e'), None, None, 767)
        ])

    def test_parse_limits_table_file(self):
        path = self.write_model_file('limits_1.tsv', '\n'.join([
            '# comment',
            '',
            'rxn_1',
            'rxn_2\t-100',
            'rxn_3\t-\t3e-1',
            'rxn_4\t-1000\t-100  # line comment'
        ]))

        limits = list(native.parse_limits_file(path))
        self.assertEqual(limits[0], ('rxn_1', None, None))
        self.assertEqual(limits[1], ('rxn_2', -100, None))
        self.assertEqual(limits[2][0], 'rxn_3')
        self.assertEqual(limits[2][1], None)
        self.assertAlmostEqual(limits[2][2], 3e-1)
        self.assertEqual(limits[3], ('rxn_4', -1000, -100))

    def test_parse_limits_table_file_too_many_fields(self):
        path = self.write_model_file('limits.tsv', '\n'.join([
            'rxn_1\t-\t100\ttext']))
        with self.assertRaises(native.ParseError):
            list(native.parse_limits_file(path))

    def test_parse_limits_yaml_file(self):
        path = self.write_model_file('limits_1.yaml', '\n'.join([
            '- include: limits_2.yml',
            '- reaction: rxn_3',
            '  lower: -1000',
            '- reaction: rxn_4',
            '  upper: 0'
        ]))

        self.write_model_file('limits_2.yml', '\n'.join([
            '- reaction: rxn_1',
            '  lower: -1',
            '  upper: 25.5',
            '- reaction: rxn_2'
        ]))

        limits = list(native.parse_limits_file(path))
        self.assertEqual(limits, [
            ('rxn_1', -1, 25.5),
            ('rxn_2', None, None),
            ('rxn_3', -1000, None),
            ('rxn_4', None, 0)
        ])

    def test_parse_model_table_file(self):
        path = self.write_model_file('model_1.tsv', '\n'.join([
            '# comment',
            'rxn_1',
            'rxn_2',
            'rxn_3',
            'rxn_4  # line comment']))

        reactions = list(native.parse_model_file(path))
        self.assertEqual(reactions, ['rxn_1', 'rxn_2', 'rxn_3', 'rxn_4'])

    def test_parse_model_yaml_file(self):
        path = self.write_model_file('model_1.yaml', '''---
            - include: model_2.yaml
            - reactions:
              - rxn_3
              - rxn_4''')

        self.write_model_file('model_2.yaml', '''---
            - groups:
              - name: First group
                reactions: [rxn_1]
              - name: Second group
                reactions: [rxn_2]''')

        reactions = list(native.parse_model_file(path))
        self.assertEqual(reactions, ['rxn_1', 'rxn_2', 'rxn_3', 'rxn_4'])


if __name__ == '__main__':
    unittest.main()



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2015  Jon Lund Steffensen <jon_steffensen@uri.edu>

import unittest

from psamm.database import DictDatabase
from psamm.metabolicmodel import MetabolicModel
from psamm.datasource.modelseed import parse_reaction
from psamm.lpsolver import generic
from psamm.gapfill import gapfind, gapfill
from psamm.reaction import Compound


class TestGapfind(unittest.TestCase):
    def setUp(self):
        self.database = DictDatabase()
        self.database.set_reaction('rxn_1', parse_reaction('=> |A|'))
        self.database.set_reaction('rxn_2', parse_reaction('|A| <=> |B|'))
        self.database.set_reaction('rxn_3', parse_reaction('|A| => |C|'))
        self.database.set_reaction('rxn_4', parse_reaction('|C| => |D|'))
        self.database.set_reaction('rxn_5', parse_reaction('|D| <=> |E|'))
        self.model = MetabolicModel.load_model(
            self.database, self.database.reactions)

        try:
            self.solver = generic.Solver(integer=True)
        except generic.RequirementsError:
            self.skipTest('Unable to find an MILP solver for tests')

    def test_gapfind(self):
        self.model.remove_reaction('rxn_4')
        compounds = set(gapfind(self.model, self.solver))
        self.assertEqual(compounds, {Compound('D'), Compound('E')})

    def test_gapfill_add_reaction(self):
        core = set(self.model.reactions) - {'rxn_4'}
        blocked = {Compound('D'), Compound('E')}

        add, rev = gapfill(self.model, core, blocked, self.solver)
        self.assertEqual(set(rev), set())
        self.assertEqual(set(add), {'rxn_4'})

    def test_gapfill_reverse_reaction(self):
        self.model.database.set_reaction('rxn_4', parse_reaction('|D| => |C|'))
        core = set(self.model.reactions)
        blocked = {Compound('D'), Compound('E')}

        add, rev = gapfill(self.model, core, blocked, self.solver)
        self.assertEqual(set(rev), {'rxn_4'})
        self.assertEqual(set(add), set())



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2015  Jon Lund Steffensen <jon_steffensen@uri.edu>

"""Definitions related to reaction equations and parsing of such equations."""

import functools


@functools.total_ordering
class Compound(object):
    """Represents a compound in a reaction equation

    A compound is a named entity in the reaction equations representing a
    chemical compound. A compound can represent a generalized chemical entity
    (e.g. polyphosphate) and the arguments can be used to instantiate a
    specific chemical entity (e.g. polyphosphate(3)) by passing a number as an
    argument or a partially specified entity by passing an expression (e.g.
    polyphosphate(n)).
    """

    def __init__(self, name, compartment=None, arguments=()):
        self._name = str(name)
        self._compartment = None if compartment is None else str(compartment)
        self._arguments = tuple(arguments)

    @property
    def name(self):
        """Name of compound"""
        return self._name

    @property
    def compartment(self):
        """Compartment of compound"""
        return self._compartment

    @property
    def arguments(self):
        """Expression argument for generalized compounds"""
        return self._arguments

    def translate(self, func):
        """Translate compound name using given function

        >>> Compound('Pb').translate(lambda x: x.lower())
        Compound('pb')
        """
        return self.__class__(
            func(self._name), self._compartment, self._arguments)

    def in_compartment(self, compartment):
        """Return an instance of this compound in the specified compartment

        >>> Compound('H+').in_compartment('e')
        Compound('H+', 'e')
        """
        return self.__class__(self._name, compartment, self._arguments)

    def __eq__(self, other):
        return (isinstance(other, Compound) and
                self._name == other._name and
                self._compartment == other._compartment and
                self._arguments == other._arguments)

    def __ne__(self, other):
        return not self == other

    def __lt__(self, other):
        if isinstance(other, Compound):
            def tuple_repr(c):
                comp = c._compartment if c._compartment is not None else ''
                return (c._name, comp, c._arguments)
            return tuple_repr(self) < tuple_repr(other)
        return NotImplemented

    def __hash__(self):
        return (hash('Compound') ^ hash(self._name) ^
                hash(self._compartment) ^ hash(self._arguments))

    def __str__(self):
        """String representation of compound

        >>> str(Compound('Phosphate'))
        'Phosphate'
        >>> str(Compound('Phosphate', 'e'))
        'Phosphate[e]'
        >>> str(Compound('Polyphosphate', None, [Expression('n')]))
        'Polyphosphate(n)'
        >>> str(Compound('Polyphosphate', 'p', [Expression('n')]))
        'Polyphosphate(n)[p]'
        """
        s = self._name
        if len(self._arguments) > 0:
            s += '({})'.format(', '.join(str(a) for a in self._arguments))
        if self._compartment is not None:
            s += '[{}]'.format(self._compartment)
        return s

    def __repr__(self):
        def str_repr(*args):
            return 'Compound({})'.format(', '.join(repr(a) for a in args))

        if len(self._arguments) == 0:
            if self._compartment is None:
                return str_repr(self._name)
            return str_repr(self._name, self._compartment)
        return str_repr(self._name, self._compartment, self._arguments)


class Reaction(object):
    """Reaction equation representation

    Each compound is associated with a stoichiometric value.
    """

    Bidir = '<=>'
    Left = '<='
    Right = '=>'

    def __init__(self, direction, left, right):
        if direction not in (Reaction.Bidir, Reaction.Left, Reaction.Right):
            raise ValueError('Invalid direction: {}'.format(direction))
        self._direction = direction
        self._left = tuple(left)
        self._right = tuple(right)

    @property
    def direction(self):
        """Direction of reaction equation"""
        return self._direction

    @property
    def left(self):
        """Compounds on the left-hand side of the reaction equation."""
        return self._left

    @property
    def right(self):
        """Compounds on the right-hand side of the reaction equation."""
        return self._right

    @property
    def compounds(self):
        """Sequence of compounds on both sides of the reaction equation

        The sign of the stoichiometric values reflect whether the compound is
        on the left-hand side (negative) or the right-hand side (positive).
        """
        return tuple((c, -v) for c, v in self._left) + self._right

    def normalized(self):
        """Return normalized reaction

        The normalized reaction will have direction Bidir or Right.
        """

        if self._direction == Reaction.Left:
            direction = Reaction.Right
            left = self._right
            right = self._left
        else:
            direction = self._direction
            left = self._left
            right = self._right

        return Reaction(direction, left, right)

    def translated_compounds(self, translate):
        """Return reaction where compound names have been translated.

        For each compound the translate function is called with the compound
        name and the returned value is used as the new compound name. A new
        reaction is returned with the substituted compound names.
        """
        left = ((compound.translate(translate), count)
                for compound, count in self._left)
        right = ((compound.translate(translate), count)
                 for compound, count in self._right)

        return Reaction(self._direction, left, right)

    def copy(self):
        """Returns a distinct copy as a new Reaction object."""
        return self.__class__(self._direction, self._left, self._right)

    def __str__(self):
        # Use the same format as ModelSEED
        def format_compound(compound, count):
            """Format compound"""
            cpdspec = str(compound)
            if count != 1:
                return '({}) |{}|'.format(count, cpdspec)
            return '|{}|'.format(cpdspec)

        def format_compound_list(cmpds):
            """Format compound list"""
            return ' + '.join(format_compound(compound, count)
                              for compound, count in cmpds)

        return '{} {} {}'.format(
            format_compound_list(self._left),
            '?' if self._direction == '' else self._direction,
            format_compound_list(self._right))

    def __repr__(self):
        return 'Reaction({}, {}, {})'.format(
            repr(self._direction), repr(self._left), repr(self._right))

    def __eq__(self, other):
        """Indicate equality of self and other"""
        return (self._direction == other._direction and
                self._left == other._left and
                self._right == other._right)

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return (hash('Reaction') ^ hash(self._direction) ^
                hash(self._left) ^ hash(self._right))


if __name__ == '__main__':
    import doctest
    doctest.testmod()
